/***************************************************************************
 *
 * communication.h - Communication package for internet game applications
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************
 *
 *    The communication package simplifies socket communication between
 * internet game servers and clients by introducing an abstraction called
 * a channel. There are two types of channel, a stream channel and a
 * datagram channel. Once two channels are connected, the same function set
 * is used to read from and write to the channel.
 *    The connection procedure is somewhat different for the two types.
 * Stream sockets are connected through a server instance. The server
 * process creates a server instance and waits for clients to arrive.
 * Clients create a channel and connects to the server. The server can
 * be polled for client connections. If a client is found, a new channel
 * is returned and the connection is established.
 *    Datagram channels aren't really connected. Instead, the local port,
 * remote port and remote Internet address are bound to the channel at each
 * side of the communication link. There are different functions available
 * to handle this binding.
 *
 ***************************************************************************/

#ifndef COMMUNICATION_H
#define COMMUNICATION_H

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>


#define COM_STREAM 0
#define COM_DGRAM  1

#define COM_CHANNEL_DATA_SIZE  32;

#define COM_POLL_NODATA   0
#define COM_POLL_DATA     1
#define COM_POLL_EXCP     2



/*
 * Channel
 *
 * An I/O channel that applications use to communicate.
 *
 */

struct comChannel
{
  int                   type;           /* COM_STREAM or COM_DGRAM for now */ 

  int                   sockfd;         /* communication socket */
  int                   localPort;      /* local port */
  int                   remotePort;     /* remote port */
  struct sockaddr_in    remoteAddr;     /* remote network address */
};
typedef struct comChannel       comChannel;


/*
 * Server
 *
 * A server that clients can connect to.
 *
 */

struct comServer
{
  int       sockfd;        /* socket */
  int       port;          /* port that clients connect to */
};
typedef struct comServer     comServer;



/*
 * comCreateServer
 *
 * Input:
 *    server
 *       Pointer to a server.
 *    port
 *       Port. If 0, the system assigns a port to the server.
 *
 * Description:
 *    Prepares the server for client connection requests.
 *
 * Returns:
 *    Port the server is listening on if successful.
 *    -1 if unsuccessful. If the error is socket-related, errno is set.
 *
 */

int comCreateServer (comServer *server, int port);


/*
 * comConnect
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *    host
 *       Name of a host.
 *    port
 *       Port of a server.
 *
 * Description:
 *    Tries to connect the channel to a server at the specified host,
 *    listening on the specified port.
 *
 * Returns:
 *    Local port of the channel if successful, -1 otherwise. Errno is set.
 *
 */

int comConnect (comChannel *channel, char *host, int port);


/*
 * comAccept
 *
 * Input:
 *    server
 *       Pointer to a server.
 *    channel
 *       Pointer to a channel.
 *
 * Description:
 *    Checks if there is a client connection request available on the
 *    server. If there is, accepts the client and initializes the channel.
 *
 * Returns:
 *    -1 if an error occurred. Errno is set.
 *    0 if no client connection requests was found.
 *    1 if successful.
 *
 */

int comAccept (comServer *server, comChannel *channel);


/*
 * comBind
 *
 * Input:
 *    channel
 *       Pointer to a DATAGRAM channel.
 *    host
 *       Name of a host.
 *    remotePort
 *       Remote port.
 *    localPort
 *       Local port.
 *
 * Description:
 *    Tries to initialize a datagram channel.
 *
 * Returns:
 *    -1 if an error occurred. Errno is set.
 *    Local port if successful.
 *
 */

int comBind (comChannel *channel, char *host, int remotePort, int localPort);


/*
 * comBindAddr
 *
 * Input:
 *    channel
 *       Pointer to a DATAGRAM channel.
 *    addr
 *       Pointer to sockaddr_in structure containing remote address.
 *    localPort
 *       Local port.
 *
 * Description:
 *    Tries to initialize a datagram channel. Note that the port of
 *    the new channel will be the same as the one in the addr structure.
 *    Typically, it is changed later with comSetRemotePort.
 *
 * Returns:
 *    -1 if an error occurred. Errno is set.
 *    Local port if successful.
 *
 */

int
comBindAddr (comChannel          *channel,
	     struct sockaddr_in  *addr,
	     int                  localPort);


/*
 * comSetRemotePort
 *
 * Input:
 *    channel
 *       Pointer to communication channel.
 *    remotePort
 *       Remote port.
 *
 * Description:
 *    This function is used to complete a datagram channel initialized
 *    by comBindAddr. The reason this function exists is because there
 *    must be a way for a client and server to exchange remote datagram
 *    ports, but also to let the system assign these ports (by binding to
 *    port 0). Thus, the typical scenario is:
 *       1) A client connects to server. A sockaddr_in structure is aquired
 *          by the server.
 *       2) The server creates a datagram channel using comBindAddr.
 *       3) The server sends its local datagram port to the client.
 *       4) The client creates a datagram channel using comBind.
 *       5) The client sends its local datagram port to the server.
 *       6) The server sets the remote datagram port by calling
 *          comSetRemotePort.
 *
 */

void comSetRemotePort (comChannel     *channel,
		       int             remotePort);


/*
 * comClose
 *
 * Input:
 *   channel
 *      Pointer to a channel.
 *
 * Description:
 *   Closes the communication channel.
 *
 * Returns:
 *   0 if successful.
 *   -1 if unsuccessful. Errno is set.
 *
 */

int comClose (comChannel *channel);


/*
 * comRead
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *    buffer
 *       Data buffer.
 *    nbytes
 *       Number of bytes to be read.
 *
 * Description:
 *    Reads a number of bytes into the specified buffer.
 *
 * Returns:
 *    Number of bytes read. -1 if an error occurred. Errno is set.
 *
 */

int comRead (comChannel *channel, void *buffer, int nbytes);


/*
 * comWrite
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *    buffer
 *       Data buffer.
 *    nbytes
 *       Number of bytes to be written.
 *
 * Description:
 *    Writes a number of bytes to the channel from the specified buffer.
 *
 * Returns:
 *    Number of bytes written. -1 if an error occurred. Errno is set.
 *
 */

int comWrite (comChannel *channel, void *buffer, int nbytes);


/*
 * comSetAddr
 *
 * Input:
 *    addr
 *       Pointer to a sockaddr_in structure to be filled in.
 *    host
 *       Host name in "127.0.0.0" format.
 *    port
 *       Port.
 *
 * Description:
 *    Fills in the sockaddr_in structure based on the data.
 *
 */

void pxSetAddr (struct sockaddr_in *addr, char *host, int port);


/*
 * comPort
 *
 * Input:
 *    sockfd
 *       Socket descriptor
 *
 * Description:
 *    Returns the port of the socket.
 *
 * Returns:
 *    Port of the socket if successful, -1 otherwise. Errno is set.
 *
 */

u_short comPort (int sockfd);


/*
 * comSetNonblocking
 *
 * Input:
 *    sockfd
 *       Socket descriptor.
 *
 * Description:
 *    Sets the socket nonblocking flag.
 *
 * Returns:
 *    -1 if unsuccessful.
 *    0 if successful.
 *
 */

int comSetNonblocking (int sockfd);


/*
 * comPrintChannel
 *
 * Input:
 *    channel
 *       Pointer to channel.
 *
 * Description:
 *    Prints channel information to stdout.
 *
 */

void comPrintChannel (comChannel   *channel);


/*
 * comPoll
 *
 * Input:
 *    channel
 *       Pointer to a communication channel.
 *
 * Description:
 *    Polls the channel for available data to read.
 *
 * Returns:
 *    COM_POLL_NODATA if there is no data available to read.
 *    COM_POLL_DATA   if there is data available to read.
 *    COM_POLL_EXCP   if there is an exception on the channel.
 *    -1 if unsuccessful. Errno is set.
 *
 */

int comPoll (comChannel    *channel);


#endif  /* COMMUNICATION_H */
