/***************************************************************************
 *
 * linkedlist.h - Functions and types that implements a linked list ADT.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#ifndef LINKEDLIST_H
#define LINKEDLIST_H


#define LL_DEBUG      /* Comment to remove debug checks */


/*
 * A entry in the linked list. This structure is for internal use only.
 *
 */

struct linkedlistentry;
struct linkedlistentry
{
  struct linkedlistentry    *next;
  struct linkedlistentry    *previous;
  void                      *data;
};
typedef struct linkedlistentry     linkedlistentry;


/*
 * A linked list.
 *
 */

struct linkedlist
{
  linkedlistentry     *first;
  int                  length;
  linkedlistentry     *cursor;
  linkedlistentry     *last;
};
typedef struct linkedlist       linkedlist;



/*
 * llNewList
 *
 * Description:
 *    Creates a new linked list.
 *
 * Returns:
 *    Pointer to the new list.
 *
 */

linkedlist *llNewList (void);


/*
 * llFirst
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the first element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that list is
 *    empty).
 *
 */

int  llFirst (linkedlist *l);


/*
 * llNext
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the next element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that cursor was
 *    pointing to last element when the call was issued).
 *
 */

int  llNext (linkedlist *l);


/*
 * llPrevious
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the previous element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that cursor was
 *    pointing to first element when the call was issued).
 *
 */

int  llPrevious (linkedlist *l);


/*
 * llLast
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the last element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that list is empty).
 *
 */

int  llLast (linkedlist *l);


/*
 * llGetData
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Returns:
 *    Pointer to data contained in the list element pointed to by the
 *    list cursor.
 *
 */

void *llGetData (linkedlist *l);


/*
 * llNewEntry
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *    data
 *       Pointer to data.
 *
 * Description:
 *    Creates a new list entry, containing the specified data. Inserts
 *    the new entry after the entry pointed to by the list cursor. The
 *    list cursor is then set to point to the new element.
 *
 * NOTE: The data is NOT copied into the element.
 *
 */

void llNewEntry (linkedlist *l, void *data);


/*
 * llDeleteEntry
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Removes the element pointed to by the list cursor.
 *
 * Returns:
 *    Pointer to the data of the deleted element.
 *
 * NOTE: The element data is NOT deallocated.
 *
 */

void *llDeleteEntry (linkedlist *l);


/*
 * llGetLength
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Returns:
 *    Number of elements in the list pointed to by l.
 *
 */

int llGetLength (linkedlist *l);

#endif /* LINKEDLIST_H */
