/***************************************************************************
 *
 * s_simulation.h - Server side simulation functions.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#ifndef S_SIMULATION_H
#define S_SIMULATION_H

#include <timeutils.h>

#include <s_client.h>
#include <s_world.h>
#include <s_collision.h>



#define SD_VESSEL_RADIUS           8.0
#define SD_VESSEL_MAXSPEED         100.0
#define SD_VESSEL_MINSPEED         0.0
#define SD_VESSEL_MAXBLASTS        10
#define SD_VESSEL_RELOADTIME       0.5
#define SD_VESSEL_WARNINGINTERVAL  1.0
#define SD_VESSEL_WARNINGS         5

#define SD_VESSEL_ROTATIONSPEED    45.0      /* Degrees / sec */
#define SD_VESSEL_THRUSTSPEED      20.0

#define SD_BLAST_RADIUS            2.0
#define SD_BLAST_SPEED             100.0
#define SD_BLAST_FUEL              200.0

#define SD_EXPLOSION_SPEED         0.1
#define SD_EXPLOSION_FRAMES        100



struct worldEntry;
struct clientConnectionData;

/* 
 * Vessel
 *
 */

struct vesselData
{
  struct clientConnectionData  *connectionData;

  vector                 position;
  vector                 heading;
  vector                 up;
  char                   animFrame;
  char                   model;

  float                  headingThrust;
  float                  pitchThrust;
  float                  rollThrust;
  float                  thrust;

  float                  speed;   /* in world coordinate units/sec */
  int                    activeBlasts;
  timestamp              lastFired;

  int                    outsideAsteroidField;
  int                    warnings;
  timestamp              lastWarning;
};
typedef struct vesselData        vesselData;


/*
 * Blast
 *
 */

struct blastData
{
  vector            position;
  vector            heading;
  vector            up;
  char              animFrame;

  float             speed;         /* in world coordinate units/sec */
  float             fuel;          /* in world units */

  int               parent;        /* ID of parent */
};
typedef struct blastData        blastData;


/*
 * Explosion
 *
 */

struct explosionData
{
  vector       position;
  char         animFrame;

  timestamp    lastFrame;
};
typedef struct explosionData      explosionData;



/* -------- VESSELS (CLIENTS) -------- */

/*
 * initVessel
 *
 * Input:
 *    vessel
 *       Pointer to vessel data.
 *    bs
 *       Pointer to a bounding sphere.
 *
 * Description:
 *    Initializes a vessel data structure.
 *
 */

void initVessel (vesselData *vessel, boundingSphere *bs);


/*
 * vesselUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to world database entry.
 *    timeslice
 *       Simulation timeslice.
 *
 * Description:
 *    Updates a vessel position and orientation based on its current
 *    state.
 *
 */

void vesselUpdateFunc (struct worldEntry *entry, double timeslice);


/*
 * getVesselXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates vessel transfer data from world entry.
 *
 */

void getVesselXferData (xferData *data, struct worldEntry *entry);


/*
 * vesselCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (vessel) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The vessel entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void vesselCollisionFunc (struct worldEntry *me, struct worldEntry *other);


/*
 * getEyePointXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates eye point transfer data from world entry.
 *
 */

void getEyePointXferData (xferData *data, struct worldEntry *entry);


/*
 * sendMsgDeath
 *
 * Input:
 *    entry
 *       World entry.
 *
 * Description:
 *    Sends a message to a client that indicates that the client's vessel
 *    has been killed.
 *
 */

void sendMsgDeath (struct worldEntry *entry);


/*
 * sendMsgScore
 *
 * Input:
 *    entry
 *       World entry.
 *
 * Description:
 *    Sends a message to a client that indicates that the client has
 *    scored a point.
 *
 */

void sendMsgScore (struct worldEntry *entry);


/*
 * sendMsgText
 *
 * Input:
 *    entry
 *       World entry.
 *    str
 *       Text string.
 *
 * Description:
 *    Sends a text message to a client.
 *
 */

void sendMsgText (struct worldEntry *entry, char *str);


/*
 * sendMsgWarning
 *
 * Input:
 *    entry
 *       World entry.
 *
 * Description:
 *    Sends a warning to a client. This is done when the client
 *    is venturing outside the asteroid field.
 *
 */

void sendMsgWarning (struct worldEntry *entry);


/*
 * updateSpeed
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel speed.
 *      
 */

void updateSpeed (vesselData *data, double timeslice);


/*
 * updateHeading
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel heading.
 *      
 */

void updateHeading (vesselData *vessel, double timeslice);


/*
 * updatePitch
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel pitch.
 *      
 */

void updatePitch (vesselData *vessel, double timeslice);


/*
 * updateRoll
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel roll.
 *      
 */

void updateRoll (vesselData *vessel, double timeslice);


/* -------- BLASTS -------- */


/*
 * ctrlFire
 *
 * Input:
 *    vessel
 *       Pointer to vessel data.
 *
 * Description:
 *    Adds a blast to the world database.
 *
 */

void ctrlFire (struct worldEntry *entry, char *data);


/*
 * blastUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to world database entry.
 *    timeslice
 *       Simulation timeslice.
 *
 * Description:
 *    Updates a blast position and orientation based on its current
 *    state.
 *
 */

void blastUpdateFunc (struct worldEntry *entry, double timeslice);


/*
 * getBlastXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates blast transfer data from world entry.
 *
 */

void getBlastXferData (xferData *data, struct worldEntry *entry);


/*
 * blastCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (blast) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The blast entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void blastCollisionFunc (struct worldEntry *me, struct worldEntry *other);


/*
 * blastDeregisterFunc
 *
 * Input:
 *    entry
 *       Pointer to (blast) world entry.
 *
 * Description:
 *    Callback for when a blast world entry is deleted.
 *
 */

void blastDeregisterFunc (struct worldEntry *entry);


/* -------- EXPLOSIONS -------- */


/*
 * explosionUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to world database entry.
 *    timeslice
 *       Simulation timeslice.
 *
 * Description:
 *    Updates an explosion based on its current state.
 *
 */

void explosionUpdateFunc (struct worldEntry *entry, double timeslice);


/*
 * getExplosionXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates explosion transfer data from world entry.
 *
 */

void getExplosionXferData (xferData *data, struct worldEntry *entry);


/*
 * explosionCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (explosion) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The explosion entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void explosionCollisionFunc (struct worldEntry *me, struct worldEntry *other);


/*
 * newExplosion
 *
 * Input:
 *    entry
 *       Pointer to vessel.
 *
 * Description:
 *    Creates an explosion at the position of the specified vessel.
 *
 */

void newExplosion (struct worldEntry    *entry);


#endif /* S_SIMULATION_H */
