/***************************************************************************
 *
 * s_world.h - Server side world database interaction functions.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#ifndef S_WORLD_H
#define S_WORLD_H

#include <transfer.h>
#include <s_simulation.h>
#include <s_client.h>
#include <s_collision.h>


/* Maximum size of a world entry */

#define SD_WORLDENTRYSIZE  256

/* World entry types */

#define SD_WE_VESSEL      0
#define SD_WE_ASTEROID    1
#define SD_WE_BLAST       2
#define SD_WE_EXPLOSION   3

/* World size */

#define SD_WORLDSIZE      1024


/*
 * Update/simulate function
 *
 */

struct worldEntry;
struct clientConnectionData;

typedef void (*updateFunction) (struct worldEntry    *entry,
				double                timeslice);

/*
 * Function that retreives transfer data
 *
 */

typedef void (*getXferDataFunction) (xferData             *data,
				     struct worldEntry    *entry);

/*
 * Function that acts on collisions.
 *
 */

typedef void (*collisionFunction) (struct worldEntry  *me,
				   struct worldEntry  *other);


/*
 * Function that is called before the entry is deleted.
 *
 */

typedef void (*deregisterFunction) (struct worldEntry *entry);


/*
 * World entry
 *
 */

struct worldEntry;
struct worldEntry
{
  int                         type;
  updateFunction              updateFunc;
  getXferDataFunction         getXferDataFunc;
  collisionFunction           collisionFunc;
  deregisterFunction          deregisterFunc;
  boundingSphere              boundingSph;
  struct worldEntry          *parent;
  
  char                        data[SD_WORLDENTRYSIZE];

  int                         deleteMark;
};
typedef struct worldEntry      worldEntry;



/*
 * newWorldEntry
 *
 * Input:
 *    type
 *       Type of new world database entry.
 *    uf
 *       Pointer to a world database entry update function. This is
 *       called during simulation.
 *    xf
 *       Pointer to a function that returns data suitable for
 *       transmission to a client.
 *    cf
 *       Pointer to function that responds to collisions.
 *    df
 *       Pointer to function that is called when object is deleted. May
 *       be NULL.
 *    bs
 *       Pointer to a bounding sphere.
 *    parent
 *       Pointer to parent. May be NULL.
 *    data
 *       Pointer to type-specific data.
 *    datasize
 *       Size of the type-specific data.
 *
 * Description:
 *    Adds a new entry to the world database.
 *
 * Returns:
 *    Pointer to the new world database entry if successful.
 *    NULL if unsuccsessful. This indicates that the world
 *    database is full.
 *
 */

worldEntry* newWorldEntry (int                         type,
			   updateFunction              uf,
			   getXferDataFunction         xf,
			   collisionFunction           cf,
			   deregisterFunction          df,
			   boundingSphere             *bs,
			   worldEntry                 *parent,
			   char                       *data,
			   int                         datasize);


/*
 * deleteClient
 *
 * Input:
 *    client
 *       Pointer to client connection data.
 *
 * Description:
 *    Removes the client connection data from the list of
 *    clients. Also removes the client world entry from the
 *    world database.
 *
 */

void deleteClient (struct clientConnectionData    *client);


/*
 * updateWorld
 *
 * Input:
 *    timeslice
 *       How long the call to updateWorld should run the simulation
 *       in seconds. I.e. when the call to updateWorld is finished,
 *       the simulation should have progressed 'timeslice' seconds.
 *
 * Description:
 *    Updates the world database by calling update functions.
 *
 */

void updateWorld (double timeslice);


/*
 * clearWorld
 *
 * Description:
 *    Initializes the world database.
 *
 */

void clearWorld (void);


/*
 * sendWorld
 *
 * Description:
 *    Send world to all clients.
 *
 */

void sendWorld (void);


/*
 * fetchVesselData
 *
 * Input:
 *    client
 *       Pointer to client connection data.
 *
 * Description:
 *    Returns a pointer to the vessel data structure that
 *    corresponds to the specified client connection data.
 *
 * Returns:
 *    Pointer to vessel data structure if successful.
 *    NULL otherwise.
 *
 */

struct vesselData;
struct vesselData *fetchVesselData (struct clientConnectionData *client);


/*
 * startPoint
 *
 * Input:
 *    position
 *       Vector to be filled in with start position.
 *    heading
 *       Vector to be filled in with start heading.
 *    up
 *       Vector to be filled in with up direction.
 *
 * Description:
 *    Calculates a new vessel/client start position based on
 *    the current vessels in the database.
 *
 */

void startPoint (vector position, vector heading, vector up);


/*
 * checkCollisions
 *
 * Description:
 *    Checks for collisions among the objects of the world.
 *
 */

void checkCollisions (void);


/*
 * findParent
 *
 * Input:
 *    entry
 *       Pointer to world entry.
 *
 * Returns:
 *    Parent of the world entry. May be NULL.
 *
 */

worldEntry *findParent (worldEntry  *entry);


#endif /* S_WORLD_H */
