/***************************************************************************
 *
 * vector.h - Vector math utilities.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#ifndef VECTOR_H
#define VECTOR_H

/* Conversion constants between degrees and radians */

#define VEC_DEG_TO_RAD 0.017453293
#define VEC_RAD_TO_DEG 57.29578


/*
 * A three-dimensional vector.
 *
 */

typedef float vector[3];

/*
 * Coefficients for the plane equation.
 *
 */

typedef float plane[4];


/*
 * vecPrint
 *
 * Input:
 *    v
 *       Vector.
 *
 * Description:
 *    Prints the vector on stdout.
 *
 */

void  vecPrint (vector v);


/*
 * vecCopy
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *
 * Description:
 *    Copies the contents of the source vector into the destination vector.
 *
 */

void  vecCopy (vector dest, vector src);


/*
 * vecLength
 *
 * Input:
 *    u
 *       Vector.
 *
 * Returns:
 *    Norm of the vector.
 *
 */

float vecLength (vector u);


/*
 * vecLengthSq
 *
 * Input:
 *    u
 *       Vector
 *
 * Returns:
 *    Square of the norm of the vector.
 *
 */

float vecLengthSq (vector u);


/*
 * vecDist
 *
 * Input:
 *    u
 *       Point.
 *    v
 *       Point.
 *
 * Returns:
 *    Geometrical distance between points u and v.
 *
 */

float vecDist (vector u, vector v);


/*
 * vecDistSq
 *
 * Input:
 *    u
 *       Point.
 *    v
 *       Point.
 *
 * Returns:
 *    Square of the geometrical distance between points u and v.
 *
 */

float vecDistSq (vector u, vector v);


/*
 * vecSub
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts (u - v) in "dest".
 * 
 */

void  vecSub (vector dest, vector u, vector v);


/*
 * vecAdd
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts (u + v) in "dest".
 * 
 */

void  vecAdd (vector dest, vector u, vector v);


/*
 * vecMult
 *
 * Input:
 *    dest
 *       Destination vector.
 *    c
 *       Scalar constant.
 *    u
 *       Vector.
 *
 * Description:
 *    Puts c * u in "dest".
 * 
 */

void  vecMult (vector dest, float c, vector u);


/*
 * vecDot
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts dot product of u and v in "dest".
 * 
 */

float vecDot (vector u, vector v);


/*
 * vecAngle
 *
 * Input:
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Returns:
 *    Geometrical angle between u and v in radians.
 * 
 */

float vecAngle (vector u, vector v);


/*
 * vecCross
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts cross product of u and v in "dest".
 * 
 */

void  vecCross (vector dest, vector u, vector v);


/*
 * vecNormalize
 *
 * Input:
 *    dest
 *       Destination vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts (v / |v|) in "dest".
 * 
 */

void  vecNormalize (vector dest, vector v);


/*
 * vecDistPointToPlane
 *
 * Input:
 *    r
 *       Point.
 *    p
 *       Plane described by plane equation.
 *
 * Returns:
 *    Geometrical distance between point and plane.
 *
 */

float vecDistPointToPlane (vector r, plane p);


/*
 * vecPlaneFromPointAndNormal
 *
 * Input:
 *    dest
 *       Destination plane.
 *    p
 *       Point on a plane.
 *    N
 *       Normal of a plane.
 *
 * Description:
 *    Calculates the plane equation corresponding to the plane
 *    described by p and N.
 *
 */

void  vecPlaneFromPointAndNormal (plane dest, vector p, vector N);


/*
 * vecSpherical
 *
 * Input:
 *    dest
 *       Destination point.
 *    theta
 *       Angle in radians.
 *    phi
 *       Angle in radians.
 *
 * Description:
 *    Calculates the vector [x y z] that corresponds to the given
 *    theta and phi in the spherical coordinates transformation
 *
 *    x = sin (theta) * cos (phi)
 *    y = sin (theta) * sin (phi)
 *    z = cos (theta)
 *
 *    where
 *
 *    0 <= theta <= PI
 *    0 <= phi <= 2*PI
 *
 */

void  vecSpherical (vector dest, float theta, float phi);


/*
 * vecRotateX
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates the source vector around the X axis.
 *
 */

void  vecRotateX (vector dest, vector src, float angle);


/*
 * vecRotateY
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates the source vector around the Y axis.
 *
 */

void  vecRotateY (vector dest, vector src, float angle);


/*
 * vecRotateZ
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates the source vector around the Z axis.
 *
 */

void  vecRotateZ (vector dest, vector src, float angle);


/*
 * vecRotate
 *
 * Input:
 *    dest
 *       Destination point.
 *    p
 *       Source point.
 *    axis
 *       Normalized direction vector that defines an axis of rotation.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates a point around an arbitrary axis.
 *
 */

void  vecRotate (vector dest, vector p, vector axis, float angle);


/*
 * vecMatch
 *
 * Input:
 *    angle
 *       Destination angle in radians.
 *    axis
 *       Destination axis.
 *    u
 *       Normalized vector.
 *    v
 *       Normalized vector.
 *
 * Description:
 *    Returns an angle R and an axis A. Rotate R radians around axis
 *    A to transform vector u to match vector v. Note that u and
 *    v must be normalized direction vectors.
 *
 */

void  vecMatch (float *angle, vector axis, vector u, vector v);


#endif /* VECTOR_H */
