/***************************************************************************
 *
 * c_asteroid.c - Client side functions for asteroid entity.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <errno.h>
#include <string.h>

#include <GL/glut.h>

#include <general.h>
#include <client.h>
#include <transfer.h>
#include <readppm.h>

#include <c_asteroid.h>
#include <c_sphere.h>
#include <s_asteroid.h>     /* for SD_ASTEROID_RADIUS */
#include <client.h>

GLuint   asteroidDL;         /* holds name of asteroid display list */
GLuint   asteroidTexture;    /* holds name of asteroid texture */


/*
 * createAsteroidTexel
 *
 * Input:
 *    texel
 *       Memory area in which to store texel data.
 *    R
 *       Red component.
 *    G
 *       Green component.
 *    B
 *       Blue component.
 *    maxColorVal
 *       (Supposedly) the maximum value any of the R, G and B components
 *       can have.
 *
 * Description:
 *    Used by the readppm package. Creates a texel from ppm R, G and B data
 *    and stores it as a texel in the memory area pointed to by "texel".
 *    Four values are stored: R, G, B and A.
 * 
 */

void
createAsteroidTexel (float   *texel,
		     int      R,
		     int      G,
		     int      B,
		     int      maxColorVal)
{
  *texel = (float) R / (float) maxColorVal;
  *(texel + 1) = (float) G / (float) maxColorVal;
  *(texel + 2) = (float) B / (float) maxColorVal;
  *(texel + 3) = 1.0;  /* Asteroids are opaque! */
}


/*
 * createAsteroidDL
 *
 * Description:
 *    Creates the asteroid display list.
 *
 */

void
createAsteroidDL (void)
{ 
  float   *texData;
  int      w, h;

  asteroidDL = glGenLists (1);

  /* Create asteroid texture */

  glGenTextures (1, &asteroidTexture);
  glBindTexture (GL_TEXTURE_2D, asteroidTexture);

  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  texData = readBinaryPPM ("asteroid.ppm",
			   createAsteroidTexel,
			   &w, &h);
  if (texData == NULL)
    {
      printf ("ERROR: Couldn't create texture from file ");
      printf ("'asteroid.ppm': %s\n", strerror (errno));
      exit (SD_ERROR);
    }
 
  printf ("[asteroid.ppm]\n");

  glTexImage2D (GL_TEXTURE_2D,
		0,
		GL_RGBA,
		w, h, 0,
		GL_RGBA,
		GL_FLOAT,
		texData);

  free (texData);

  /* Asteroid */

  glNewList (asteroidDL, GL_COMPILE);

  drawSphere (1.0, 4, 4);

  glEndList ();
}


/*
 * drawAsteroid
 *
 * Input:
 *    data
 *       Pointer to transfer data containing the asteroid data.
 *    eye
 *       Pointer to eyepoint data.
 *
 * Description:
 *    Renders an asteroid.
 *
 */

void
drawAsteroid (xferData           *data,
	      eyePointXferData   *eye)
{
  asteroidXferData   asteroid;
  float              amb[4] = { 0.2, 0.15, 0.11, 1.0 };
  float              dif[4] = { 0.3, 0.1, 0.0, 1.0 };
  float              spc[4] = { 0.0, 0.0, 0.0, 1.0 };
 
  float              tamb[4] = { 0.2, 0.2, 0.2, 1.0 };
  float              tdif[4] = { 0.6, 0.6, 0.6, 1.0 };
  float              tspc[4] = { 0.0, 0.0, 0.0, 1.0 };


  memcpy ((char *) &asteroid, (char *) &data->data, sizeof (asteroid));

  if (nearEnough (asteroid.position, eye))
    {
      if (pointVisible (asteroid.position, eye))
	{
	  glPushMatrix ();

	  glTranslatef (asteroid.position[0],
			asteroid.position[1],
			asteroid.position[2]);
	  glRotatef (asteroid.angleX, 1.0, 0.0, 0.0);
	  glRotatef (asteroid.angleY, 0.0, 1.0, 0.0);
	  glRotatef (asteroid.angleZ, 0.0, 0.0, 1.0);
	  glScalef (asteroid.radius, asteroid.radius, asteroid.radius);

	  glEnable (GL_DEPTH_TEST);
	  glEnable (GL_LIGHTING);
	  glEnable (GL_NORMALIZE);
	  glShadeModel (GL_SMOOTH);

	  if (displayOptions.wantTextures)
	    {
	      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, tamb);
	      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, tdif);
	      glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, tspc);

	      glEnable (GL_TEXTURE_2D);
	      glBindTexture (GL_TEXTURE_2D, asteroidTexture);
	      
	      if (displayOptions.wantAsteroidLighting)
		{
		  glTexEnvi (GL_TEXTURE_ENV,
			     GL_TEXTURE_ENV_MODE, 
			     GL_MODULATE);
		}
	      else
		{
		  glTexEnvi (GL_TEXTURE_ENV,
			     GL_TEXTURE_ENV_MODE, 
			     GL_REPLACE);
		}
	    }
	  else
	    {
	      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, amb);
	      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, dif);
	      glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, spc);
	    }

	  glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 0.0);

	  glCallList (asteroidDL);

	  glPopMatrix ();
	}
    }
}

