/***************************************************************************
 *
 * c_radar.c - Client-side functions for drawing the radar screens.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/


#include <GL/glut.h>
#include <math.h>

#include <general.h>
#include <vector.h>
#include <transfer.h>
#include <client.h>
#include <c_radar.h>
#include <c_vessel.h>
#include <c_blast.h>
#include <c_explosion.h>


GLuint   radarDL;       /* holds the name of the radar dot display list */


/*
 * createRadarDL
 *
 * Description:
 *    Creates the radar dot display list.
 *
 */

void
createRadarDL (void)
{
  radarDL = glGenLists (1);
  
  glNewList (radarDL, GL_COMPILE);

  glBegin (GL_QUADS);
  glVertex2f (-1.5, -1.5);
  glVertex2f ( 1.5, -1.5);
  glVertex2f ( 1.5,  1.5);
  glVertex2f (-1.5,  1.5);
  glEnd ();

  glEndList ();
}


/*
 * drawRadarSquares
 *
 * Input:
 *    radarWidth
 *       Width of the radar.
 *    radarHeight
 *       Height of the radar.
 *    win_width
 *       Client window width.
 *    win_height
 *       Client window height.
 *
 * Description:
 *    Draws two green radar squares.
 *
 */

void
drawRadarSquares (GLfloat   radarWidth,
		  GLfloat   radarHeight,
		  GLfloat   win_width,
		  GLfloat   win_height)
{
  glDisable (GL_LIGHTING);
  glColor4f (0.3, 0.6, 0.3, 0.5);

#define SD_RADARBLEND
#ifdef SD_RADARBLEND
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable (GL_BLEND);
#endif

  glViewport (0.0, 0.0, floor (radarWidth), floor (radarHeight));
  glBegin (GL_QUADS);
  glVertex2f (0.0, 0.0);
  glVertex2f (radarWidth, 0.0);
  glVertex2f (radarWidth, radarHeight);
  glVertex2f (0.0, radarHeight);
  glEnd ();

  glViewport (floor (win_width - radarWidth), 1.0, floor (radarWidth), 
	      floor (radarHeight));
  glBegin (GL_QUADS);
  glVertex2f (0.0, 0.0);
  glVertex2f (radarWidth, 0.0);
  glVertex2f (radarWidth, radarHeight);
  glVertex2f (0.0, radarHeight);
  glEnd ();
}


/*
 * radarInverseRotation
 *
 * Input:
 *    dest
 *       Destination position.
 *    src
 *       Source position.
 *    heading
 *       Heading vector.
 *    up
 *       Up vector.
 *
 * Description:
 *    Transforms the input position to match the heading and up vectors.
 *
 * BUGS:
 *    dest[0] and dest[2] must be negated to get the expected result.
 *
 */

void
radarInverseRotation (vector dest,
		      vector src,
		      vector heading,
		      vector up)
{
  vector    x;
  vector    y;
  vector    z;

  vecCopy (z, heading);
  vecNormalize (z, z);
  vecCopy (y, up);
  vecCross (x, y, z);
  vecCross (y, z, x);
  vecNormalize (x, x);
  vecNormalize (y, y);

  dest[0] = src[0] * x[0] + src[1] * x[1] + src[2] * x[2];
  dest[1] = src[0] * y[0] + src[1] * y[1] + src[2] * y[2];
  dest[2] = src[0] * z[0] + src[1] * z[1] + src[2] * z[2];
}


/*
 * drawRadarVessel
 *
 * Draw a radar vessel.
 *
 */

void
drawRadarVessel (xferData    *data,
		 vector       position,
		 vector       heading,
		 vector       up,
		 float        rw,
		 float        rh,
		 float        ww,
		 float        wh)
{
  vesselXferData     vessel;
  vector             d;
  vector             d1;
  float              x;
  float              y;

  memcpy ((char *) &vessel, (char *) data->data, sizeof (vessel));

  vecSub (d, vessel.position, position);
  vecNormalize (d, d);
  radarInverseRotation (d1, d, heading, up);
  d1[0] = -d1[0];
  d1[2] = -d1[2];

  /* Draw radar point */

  glPushMatrix ();

  if (d1[2] < 0)
    {
      glViewport (0.0, 0.0, rw, rh);
    }
  else
    {
      glViewport (ww - rw, 0.0, rw, rh);
    }

  x = (rw * 0.5 * d1[0]) + (rw / 2);
  y = (rh * 0.5 * d1[1]) + (rh / 2);
  glTranslatef (x, y, 0.0);

  glColor3f (1.0, 0.0, 0.0);
  glCallList (radarDL);

  glPopMatrix ();
}


/*
 * drawRadarVessel
 *
 * Draw a radar blast.
 *
 */

void
drawRadarBlast (xferData    *data,
		vector       position,
		vector       heading,
		vector       up,
		float        rw,
		float        rh,
		float        ww,
		float        wh)
{
  blastXferData      blast;
  vector             d;
  vector             d1;
  float              x;
  float              y;

  memcpy ((char *) &blast, (char *) data->data, sizeof (blast));

  vecSub (d, blast.position, position);
  vecNormalize (d, d);
  radarInverseRotation (d1, d, heading, up);
  d1[0] = -d1[0];
  d1[2] = -d1[2];

  /* Draw radar point */

  glPushMatrix ();
  if (d1[2] < 0)
    {
      glViewport (0.0, 0.0, rw, rh);
    }
  else
    {
      glViewport (ww - rw, 0.0, rw, rh);
    }

  x = (rw * 0.5 * d1[0]) + (rw / 2);
  y = (rh * 0.5 * d1[1]) + (rh / 2);
  glTranslatef (x, y, 0.0);

  glColor3f (1.0, 1.0, 0.0);
  glCallList (radarDL);

  glPopMatrix ();
}


/*
 * drawRadarVessel
 *
 * Draw a radar asteroid.
 *
 */

void
drawRadarAsteroid (xferData    *data,
		   vector       position,
		   vector       heading,
		   vector       up,
		   float        rw,
		   float        rh,
		   float        ww,
		   float        wh)
{
  asteroidXferData   asteroid;
  vector             d;
  vector             d1;
  float              x;
  float              y;

  memcpy ((char *) &asteroid, (char *) data->data, sizeof (asteroid));

  vecSub (d, asteroid.position, position);
  vecNormalize (d, d);
  radarInverseRotation (d1, d, heading, up);
  d1[0] = -d1[0];
  d1[2] = -d1[2];

  /* Draw radar point */

  glPushMatrix ();

  if (d1[2] < 0)
    {
      glViewport (0.0, 0.0, rw, rh);
    }
  else
    {
      glViewport (ww - rw, 0.0, rw, rh);
    }

  x = (rw * 0.5 * d1[0]) + (rw / 2);
  y = (rh * 0.5 * d1[1]) + (rh / 2);
  glTranslatef (x, y, 0.0);

  glColor3f (0.5, 0.5, 0.5);
  glCallList (radarDL);

  glPopMatrix ();
}


/*
 * drawRadar
 *
 * Input:
 *    datagram
 *       Array containing transfer data from server.
 *    position
 *       Position of client's vessel.
 *    heading
 *       Heading of client's vessel.
 *    up
 *       Up direction of client's vessel.
 *
 * Description:
 *    Draws the radar screens.
 *
 */

void
drawRadar (xferData   datagram[],
	   vector     position,
	   vector     heading,
	   vector     up)
{
  float            radarWidth;
  float            radarHeight;
  float            win_width;
  float            win_height;
  int              i;

  getWinDimensions (&win_width, &win_height);

  /* Radar dimensions are 1/4:th of window size */

  radarWidth = (float) win_width / 8.0;
  radarHeight = (float) win_height / 8.0;

  /* Set radar viewport */

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
	  
  gluOrtho2D (-0.01, radarWidth + 0.01, 0.0, radarHeight);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  
  drawRadarSquares (radarWidth, radarHeight, win_width, win_height);
  glClear (GL_DEPTH_BUFFER_BIT);
  
  /* Draw radar points */

  i = 0;
  while (datagram[i].type != SD_EOD)
    {      
      switch (datagram[i].type)
	{
	case SD_VESSEL:
	  drawRadarVessel (&datagram[i], 
			   position,
			   heading,
			   up,
			   radarWidth,
			   radarHeight,
			   win_width,
			   win_height);
	  break;

	case SD_BLAST:
	  drawRadarBlast (&datagram[i], 
			  position,
			  heading,
			  up,
			  radarWidth,
			  radarHeight,
			  win_width,
			  win_height);
	  break;

	case SD_ASTEROID:
	  drawRadarAsteroid (&datagram[i], 
			     position,
			     heading,
			     up,
			     radarWidth,
			     radarHeight,
			     win_width,
			     win_height);
	  break;

	default:
	  break;
	}

      ++i;
    }
}


