/***************************************************************************
 *
 * c_starfield.c - Client-side functions that draws a starfield.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <time.h>

#include <vector.h>
#include <readppm.h>
#include <randomutils.h>

#include <general.h>
#include <c_starfield.h>

GLint  starfieldDL;         /* holds the name of the starfield display list */
GLuint starfieldTexture[5]; /* holds the names of the starfield textures */


/*
 * drawRectangleD
 *
 * Input:
 *    4 corners.
 *
 * Description:
 *    Draw a rectangle with texture coordinates. The texture is
 *    repeated twice in the X and Y directions.
 *
 */

void
drawRectangleD (GLfloat    c1x,
		GLfloat    c1y,
		GLfloat    c1z,
		GLfloat    c2x,
		GLfloat    c2y,
		GLfloat    c2z,
		GLfloat    c3x,
		GLfloat    c3y,
		GLfloat    c3z,
		GLfloat    c4x,
		GLfloat    c4y,
		GLfloat    c4z)
{
  glTexCoord2f (0.0, 0.0);
  glVertex3f (c1x, c1y, c1z);

  glTexCoord2f (2.0, 0.0);
  glVertex3f (c2x, c2y, c2z);

  glTexCoord2f (2.0, 2.0);
  glVertex3f (c3x, c3y, c3z);

  glTexCoord2f (0.0, 2.0);
  glVertex3f (c4x, c4y, c4z);
}


/*
 * drawRectangle
 *
 * Input:
 *    4 corners.
 *
 * Description:
 *    Draw a rectangle with texture coordinates.
 *
 */

void
drawRectangle (GLfloat    c1x,
	       GLfloat    c1y,
	       GLfloat    c1z,
	       GLfloat    c2x,
	       GLfloat    c2y,
	       GLfloat    c2z,
	       GLfloat    c3x,
	       GLfloat    c3y,
	       GLfloat    c3z,
	       GLfloat    c4x,
	       GLfloat    c4y,
	       GLfloat    c4z)
{
  glTexCoord2f (0.0, 0.0);
  glVertex3f (c1x, c1y, c1z);

  glTexCoord2f (1.0, 0.0);
  glVertex3f (c2x, c2y, c2z);

  glTexCoord2f (1.0, 1.0);
  glVertex3f (c3x, c3y, c3z);

  glTexCoord2f (0.0, 1.0);
  glVertex3f (c4x, c4y, c4z);
}


/*
 * createStarfieldTexel
 *
 * Description:
 *    Used by readPPM. Creates texel from image data.
 *
 */

void
createStarfieldTexel (float   *texel,
		      int      R,
		      int      G,
		      int      B,
		      int      maxc)
{
  *texel = (float) R / (float) maxc;
  *(texel + 1) = (float) G / (float) maxc;
  *(texel + 2) = (float) B / (float) maxc;
  *(texel + 3) = 1.0;
}

/*
 * createStarfieldTexture
 *
 * Input:
 *    textureNo
 *       OpenGL texture name.
 *    filename
 *       Name of file containing texture data.
 *
 * Description:
 *    Read texture data and create a texture.
 *
 */

void
createStarfieldTexture (GLuint textureNo, char *filename)
{
  float            *textureData;
  int               w, h;

  textureData = readBinaryPPM (filename, createStarfieldTexel, &w, &h);

  if (textureData == NULL)
    {
      printf ("ERROR: Couldn't create texture from file ");
      printf ("'%s': %s\n", filename, strerror (errno));
      exit (SD_ERROR);
    }

  printf ("[%s]\n", filename);

  glBindTexture (GL_TEXTURE_2D, textureNo);

  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexEnvi (GL_TEXTURE_2D, GL_TEXTURE_ENV, GL_REPLACE);

  glTexImage2D (GL_TEXTURE_2D,
		0, 
		GL_RGBA,
		w, h, 0,
		GL_RGBA,
		GL_FLOAT,
		textureData);

  free (textureData);
}


/*
 * makeTexturedStarfield
 *
 * Input:
 *    dl
 *       OpenGL display list name.
 *
 * Description:
 *    Creates a starfield display list using textures.
 *
 */

void
makeTexturedStarfield (GLuint dl)
{
  glGenTextures (5, starfieldTexture);

  createStarfieldTexture (starfieldTexture[0], "stars.ppm"); 
  createStarfieldTexture (starfieldTexture[1], "starbox1.ppm");
  createStarfieldTexture (starfieldTexture[2], "starbox2.ppm");
  createStarfieldTexture (starfieldTexture[3], "starbox3.ppm");
  createStarfieldTexture (starfieldTexture[4], "starbox4.ppm");

  glNewList (dl, GL_COMPILE);

  glDisable (GL_DEPTH_TEST);
  glEnable (GL_CULL_FACE);

  glEnable (GL_TEXTURE_2D);

  /* starbox */

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[3]);
  glBegin (GL_QUADS);
  drawRectangle (-1.0,  0.0, -1.0,
		  0.0,  0.0, -1.0,
		  0.0,  1.0, -1.0,
		 -1.0,  1.0, -1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[4]);
  glBegin (GL_QUADS);
  drawRectangle ( 0.0,  0.0, -1.0,
		  1.0,  0.0, -1.0,
		  1.0,  1.0, -1.0,
		  0.0,  1.0, -1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[1]);
  glBegin (GL_QUADS);
  drawRectangle (-1.0, -1.0, -1.0,
		  0.0, -1.0, -1.0,
		  0.0,  0.0, -1.0,
		 -1.0,  0.0, -1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[2]);
  glBegin (GL_QUADS);
  drawRectangle ( 0.0, -1.0, -1.0,
		  1.0, -1.0, -1.0,
		  1.0,  0.0, -1.0,
		  0.0,  0.0, -1.0);
  glEnd ();

  /* rest of starfield */

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[0]);
  glBegin (GL_QUADS);
  drawRectangleD (1.0, -1.0, -1.0,
		 1.0, -1.0, 1.0,
		 1.0, 1.0, 1.0,
		 1.0, 1.0, -1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[0]);
  glBegin (GL_QUADS);
  drawRectangleD (1.0, -1.0, 1.0,
		 -1.0, -1.0, 1.0,
		 -1.0, 1.0, 1.0,
		 1.0, 1.0, 1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[0]);
  glBegin (GL_QUADS);
  drawRectangleD (-1.0, -1.0, 1.0,
		 -1.0, -1.0, -1.0,
		 -1.0, 1.0, -1.0,
		 -1.0, 1.0, 1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[0]);
  glBegin (GL_QUADS);
  drawRectangleD (1.0, 1.0, 1.0,
		 -1.0, 1.0, 1.0,
		 -1.0, 1.0, -1.0,
		 1.0, 1.0, -1.0);
  glEnd ();

  glBindTexture (GL_TEXTURE_2D, starfieldTexture[0]);
  glBegin (GL_QUADS);
  drawRectangleD (1.0, -1.0, -1.0,
		 -1.0, -1.0, -1.0,
		 -1.0, -1.0, 1.0,
		 1.0, -1.0, 1.0);
  glEnd ();

  glDisable (GL_TEXTURE_2D);

  glEndList ();
}



/*
 * starfieldCreateStar
 *
 * Description:
 *    Draws a star.
 *
 */

void
starfieldCreateStar (void)
{
  GLfloat r, g, b;
  GLfloat x, y, z;
  GLfloat len;

  r = 1.0 - (randomBetweenZeroAndOne () * 0.2);
  g = 1.0 - (randomBetweenZeroAndOne () * 0.2);
  b = 1.0 - (randomBetweenZeroAndOne () * 0.2);
  glColor3f (r, g, b);

  x = randomBetweenMinusOneAndOne ();
  y = randomBetweenMinusOneAndOne ();
  z = randomBetweenMinusOneAndOne ();

  len = sqrt (x * x + y * y + z * z);
  x = SD_STARFIELD_RADIUS * (x / len);
  y = SD_STARFIELD_RADIUS * (y / len);
  z = SD_STARFIELD_RADIUS * (z / len);

  glVertex3f (x, y, z);
}


/*
 * createStarfieldDL
 *
 * Input:
 *    wantTexture
 *       If true, use textures for starfield.
 *
 * Description:
 *    Creates the starfield display list.
 *
 */

void
createStarfieldDL (int wantTexture)
{
  int   i;

  starfieldDL = glGenLists (1);
      
  if (!wantTexture)
    {
      glNewList (starfieldDL, GL_COMPILE);

      glBegin (GL_POINTS);
      for (i = 0; i < SD_STARFIELD_NUM_STARS; i++)
	{
	  starfieldCreateStar ();
	}
      glEnd ();

      glEndList ();
    }
  else
    {
      makeTexturedStarfield (starfieldDL);
    }
}


/*
 * drawStarfield
 *
 * Description:
 *    Draws the starfield.
 *
 */

void
drawStarfield (void)
{
  glPushMatrix ();
  glCallList (starfieldDL);
  glPopMatrix ();
}
