/***************************************************************************
 *
 * client.c - SpaceDuel3D client
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include <GL/glut.h>

#include <general.h>
#include <communication.h>
#include <transfer.h>
#include <timeutils.h>

#include <client.h>
#include <c_starfield.h>
#include <c_blast.h>
#include <c_vessel.h>
#include <c_explosion.h>
#include <c_asteroid.h>
#include <c_radar.h>


comChannel  streamChn;    /* Stream channel to server */
comChannel  dgramChn;     /* Datagram channel to server */

xferData    datagram[SD_DATALENGTH];

controlKeys controls;
int         mouseCtrl = 0;            /* if true, use mouse to steer */

int         win_width = 320;
int         win_height = 200;
int         mouse_x = 0;              /* mouse cursor position (x) */
int         mouse_y = 0;              /* mouse cursor position (y) */

int         deathFrame = 0;           /* Death animation frame counter */
int         warningFrame = 0;         /* Warning animation frame counter */

int         score = 0;

int         framesDrawn = 0;
timestamp   startTime;                /* Time client was started */

GLuint      resetDisplayList = 1;     /* holds name of reset display list */

displayOps  displayOptions;



void connectToServer (char *host, int port);
void display (void);
void reshape (int w, int h);
void idle (void);
void getEyePoint (eyePointXferData *eye);
void drawObjects (eyePointXferData *eye);
void readWorldData (void);
void initControls (void);
void keyboard (unsigned char key, int x, int y);
void mouse (int x, int y);
void mousePress (int button, int state, int x, int y);
void sendMouseCtrl (void);
void init (void);
void createResetDisplayList (void);
void drawInfo (float speed, int score, float w, float h);
void drawMouse (float w, float h, float mx, float my);
void printTextMessage (message *msg);
void checkOptions (int argc, char *argv[]);
void initDisplayOptions (void);


int
main (int     argc,
      char   *argv[])
{
  int         streamPort;

  printf ("================================================\n");
  printf ("SpaceDuel3D Client\n\n");
  printf ("Version 1.0 BETA\n");
  printf ("Copyright (C) 1998 Gustav Taxen.\n");
  printf ("------------------------------------------------\n");
  printf ("SpaceDuel3D comes with ABSOLUTELY NO WARRANTY.\n");
  printf ("This is free software, and you are welcome\n");
  printf ("to redistribute it under certain conditions.\n");
  printf ("Please read the file 'COPYING' in the root\n");
  printf ("directory of the SpaceDuel3D distribution.\n");
  printf ("================================================\n\n");

  /* Check command line arguments */

  if (argc < 3)
    {
      printf ("Usage: %s host port [options]\n\n", argv[0]);
      printf ("Options:\n");
      printf ("  [-notextures]  Don't use asteroid textures.\n");
      printf ("  [-dithering]   Use dithering.\n");
      printf ("  [-starfield]   Use textured starfield.\n");
      printf ("  [-ambient a]   Set global ambient light factor (0.0-1.0).\n");
      printf ("  [-2]           Set window size to 640x480.\n");
      return SD_ERROR;
    }

  initDisplayOptions ();
  checkOptions (argc, argv);

  /* Initialize OpenGL */

  printf ("Initializing OpenGL...\n");
  glutInit (&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize (win_width, win_height);
  glutCreateWindow ("SpaceDuel3D");
  glutDisplayFunc (display);
  glutReshapeFunc (reshape);
  glutIdleFunc (idle);
  glutKeyboardFunc (keyboard);
  glutMouseFunc (mousePress);

  /* Initialize world database */

  printf ("Creating textures and display lists...\n");
  init ();
  
  /* Connect to server */

  streamPort = atoi (argv[2]);
  printf ("Trying %s:%d ...\n", argv[1], streamPort);
  connectToServer (argv[1], streamPort);

  /* Leave control to GLUT */
  
  startTime = getTimestamp ();
  glutMainLoop ();

  return SD_OK;   /* ANSI C requires that main returns an int */
}


/*
 * connectToServer
 *
 * Input:
 *    host
 *       Remote host name.
 *    port
 *       Remote host port.
 *
 * Description:
 *    Attempts to connect to the specified host and port.
 *
 */

void
connectToServer (char    *host,
		 int      port)
{
  int         dgramPort;
  int         localPort;

  if (comConnect (&streamChn, host, port) < 0)
    {
      printf ("ERROR: %s\n", strerror (errno));
      exit (SD_ERROR);
    }

  if (sendHandshake (&streamChn) == SD_ERROR)
    {
      comClose (&streamChn);
      printf ("ERROR: Refused by server\n");
      exit (SD_ERROR);
    }

  if (readHandshake (&streamChn) == SD_ERROR)
    {
      comClose (&streamChn);
      printf ("ERROR: Wrong server version\n");
      exit (SD_ERROR);
    }

  printf ("Connected.\n");

  /* Create datagram channel */

  comRead (&streamChn, &dgramPort, sizeof (dgramPort));
  printf ("Remote datagram port is %d.\n", dgramPort);

  localPort = comBind (&dgramChn, host, dgramPort, 0);
  printf ("Local datagram port is %d.\n", localPort);
  comWrite (&streamChn, &localPort, sizeof (localPort));  
}


/*
 * display
 *
 * Description:
 *    Repaints the OpenGL window.
 *
 */

void
display (void)
{
  eyePointXferData    eye;

  if (warningFrame)
    {
      glClearColor ((float) warningFrame / (float) SD_WARNINGFRAMES,
		    0.0,
		    0.0,
		    0.0);
      --warningFrame;
    }
  else
    {
      glClearColor (0.0, 0.0, 0.0, 0.0);
    }

  if (deathFrame)
    {
      glClearColor ((float) deathFrame / (float) SD_DEATHFRAMES,
		    (float) deathFrame / (float) SD_DEATHFRAMES,
		    (float) deathFrame / (float) SD_DEATHFRAMES,
		    0.0);
      --deathFrame;
    }

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glViewport (0.0, 0.0, (GLfloat) win_width, (GLfloat) win_height);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (SD_FOV, 
		  (GLfloat) win_width / (GLfloat) win_height,
		  0.1, 2.0);
  glMatrixMode (GL_MODELVIEW);

  getEyePoint (&eye);
  
  /* Draw starfield */

  resetOpenGL ();

  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, 
		  displayOptions.globalAmbientFactor);  

  glDisable (GL_DEPTH_TEST);
  glLoadIdentity ();
  gluLookAt (0.0, 0.0, 0.0,
	     eye.heading[0], eye.heading[1], eye.heading[2],
	     eye.up[0], eye.up[1], eye.up[2]);

  if (!(displayOptions.wantTexturedStarfield 
	&& (warningFrame || deathFrame)))
    {
      drawStarfield ();
    }

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (SD_FOV, 
		  (GLfloat) win_width / (GLfloat) win_height,
		  SD_FRONT_CLIP, SD_BACK_CLIP);
  glMatrixMode (GL_MODELVIEW);

  /* Place camera */

  glLoadIdentity ();
  gluLookAt (eye.position[0], 
	     eye.position[1],
	     eye.position[2],
	     eye.position[0] + eye.heading[0],
	     eye.position[1] + eye.heading[1],
	     eye.position[2] + eye.heading[2],
	     eye.up[0],
	     eye.up[1],
	     eye.up[2]);
  
  glClear (GL_DEPTH_BUFFER_BIT);
  glPushMatrix ();
  drawObjects (&eye);
  glPopMatrix ();

  glClear (GL_DEPTH_BUFFER_BIT);
  glPushMatrix ();
  drawRadar (datagram, eye.position, eye.heading, eye.up);
  glPopMatrix ();
  
  glPushMatrix ();
  drawInfo (eye.speed, score, (float) win_width, (float) win_height);
  glPopMatrix ();

  if (mouseCtrl)
    {
      drawMouse ((float) win_width, 
		 (float) win_height,
		 (float) mouse_x,
		 (float) mouse_y);
    }

  glutSwapBuffers ();
  ++framesDrawn;
}


/*
 * createResetDisplayList
 *
 * Description:
 *    Creates a display list that resets OpenGL parameters to
 *    minimal (= max performance).
 *
 */

void
createResetDisplayList (void)
{
  resetDisplayList = glGenLists (1);

  glNewList (resetDisplayList, GL_COMPILE);

  /*  glDisable (GL_DEPTH_TEST);*/
  glEnable (GL_DEPTH_TEST);
  glDisable (GL_ALPHA_TEST);
  glDisable (GL_STENCIL_TEST);

  glDisable (GL_LIGHTING);
  glShadeModel (GL_FLAT);

  glDisable (GL_BLEND);

  glDisable (GL_NORMALIZE);

  glDisable (GL_TEXTURE_2D);

  glEnable (GL_CULL_FACE);
  
  glColor3f (1.0, 1.0, 1.0);

  glEndList ();
}


/*
 * resetOpenGL
 *
 * Description:
 *    Reset OpenGL parameters to minimal (= max performance).
 *
 */

void
resetOpenGL (void)
{
  glCallList (resetDisplayList);
  
  if (!displayOptions.wantDithering)
    {
      glDisable (GL_DITHER);
    }
}


/*
 * reshape
 *
 * Input:
 *    w
 *       Window width.
 *    h
 *       Window height.
 *
 * Description:
 *    Resets the viewport and perspective based on the new window
 *    dimensions.
 *
 */

void
reshape (GLint     w,
	 GLint     h)
{
  win_width = w;
  win_height = h;
  mouse_x = (float) w / 2.0;
  mouse_y = (float) h / 2.0;
}


/*
 * idle
 *
 * Description:
 *    Read world database from server.
 *
 */

void
idle (void)
{
  readWorldData ();
  
  if (mouseCtrl)
    {
      sendMouseCtrl ();
    }

  glutPostRedisplay ();
}


/*
 * readWorldData
 *
 * Description:
 *    Reads datagram containing world data from server.
 *
 */

void
readWorldData (void)
{
  message   msg;

  while (comPoll (&dgramChn) == COM_POLL_DATA)
    {
      /* TODO: datagram sequence numbers */
      comRead (&dgramChn, datagram, SD_DATALENGTH * sizeof (xferData));
    }

  while (comPoll (&streamChn) == COM_POLL_DATA)
    {
      comRead (&streamChn, &msg, sizeof (msg));

      switch (msg.type)
	{
	case SD_MSG_DEATH:
	  deathFrame = SD_DEATHFRAMES;
	  break;
	case SD_MSG_SCORE:
	  ++score;
	  break;
	case SD_MSG_TEXT:
	  printTextMessage (&msg);
	  break;
	case SD_MSG_WARNING:
	  warningFrame = SD_WARNINGFRAMES;
	  break;
	default:
	  printf ("ERROR: Unknown message: %d\n", msg.type);
	  exit (SD_ERROR);
	  break;
	}
    }
}


/*
 * getEyePoint
 *
 * Input:
 *    eye
 *       Pointer to structure to fill in.
 *
 * Description:
 *    Fetches the eye point information from the datagram sent by
 *    the server.
 *
 */

void
getEyePoint (eyePointXferData   *eye)
{
  int         i;

  assert (eye != NULL);

  i = 0;
  while (datagram[i].type != SD_EOD)
    {
      if (datagram[i].type == SD_EYEPOINT)
	{
	  memcpy ((char *) eye,
		  (char *) &(datagram[i].data), 
		  sizeof (eyePointXferData));
	  return;
	}
      ++i;
    }    

  /* No eyepoint data is OK, server may not have started sending yet */
}


/*
 * drawObjects
 *
 * Input:
 *    eye
 *       Pointer to eye point data.
 *
 * Description:
 *    Steps through the transferred data and draws all objects.
 *
 */

void
drawObjects (eyePointXferData *eye)
{
  int  i;

  i = 0;
  while (datagram[i].type != SD_EOD)
    {
      switch (datagram[i].type)
	{
	case SD_EYEPOINT:
	  break;
	  
	case SD_VESSEL:
	  drawVessel (&datagram[i], eye);
	  break;

	case SD_BLAST:
	  drawBlast (&datagram[i], eye);
	  break;

	case SD_EXPLOSION:
	  drawExplosion (&datagram[i], eye);
	  break;

	case SD_ASTEROID:
	  drawAsteroid (&datagram[i], eye);
	  break;

	default:
	  printf ("ERROR: Unkown transfer data type: %d\n",
		  datagram[i].type);
	  exit (SD_ERROR);
	}
      resetOpenGL ();
      ++i;
    }
}


/*
 * initControls
 *
 * Description:
 *    Initializes the keyboard configuration.
 *
 */

void
initControls (void)
{
  controls.thrustIncr = 'a';
  controls.thrustDecr = 'z';
  controls.headingIncr = 's';
  controls.headingDecr = 'x';
  controls.pitchIncr = 'c';
  controls.pitchDecr = 'd';
  controls.rollIncr = 'f';
  controls.rollDecr = 'v';
  controls.fire = ' ';
}


/*
 * keyboard
 *
 * Input:
 *    key
 *       Keypress.
 *    x
 *       Mouse cursor x.
 *    y
 *       Mouse cursor y.
 *
 * Description:
 *    Keyboard callback function. If the keypress corresponds to a key
 *    in the configuration, the corresponding control constant is sent
 *    to the server.
 *
 */

void
keyboard (unsigned char     key,
	  int               x,
	  int               y)
{
  message                   msg;
  double                    elapsed;
  
  if (key == 27)
    {
     
      elapsed = timeDifference (startTime);
      printf ("\nFrames drawn:        %d\n", framesDrawn);
      printf ("Uptime:              %f seconds\n", elapsed);
      printf ("Estimated framerate: %f frames/sec\n", 
	      (double) framesDrawn / elapsed);

      printf ("Exiting.\n");
      exit (SD_OK);
    }

  if (key == controls.thrustIncr)
    {
      msg.type = SD_MSG_THRUST;
      *((float *) &msg.data) = 1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));
    }

  if (key == controls.thrustDecr)
    {
      msg.type = SD_MSG_THRUST;
      *((float *) &msg.data) = -1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));
    }

  if (key == controls.headingIncr)
    {
      msg.type = SD_MSG_HEADING;
      *((float *) &msg.data) = 1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));
    }

  if (key == controls.headingDecr)
    {
      msg.type = SD_MSG_HEADING;
      *((float *) &msg.data) = -1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));
    }

  if (key == controls.pitchIncr)
    {
      msg.type = SD_MSG_PITCH;
      *((float *) &msg.data) = 1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));      
    }

  if (key == controls.pitchDecr)
    {
      msg.type = SD_MSG_PITCH;
      *((float *) &msg.data) = -1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));      
    }

  if (key == controls.rollIncr)
    {
      msg.type = SD_MSG_ROLL;
      *((float *) &msg.data) = 1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));      
    }

  if (key == controls.rollDecr)
    {
      msg.type = SD_MSG_ROLL;
      *((float *) &msg.data) = -1.0;
      comWrite (&dgramChn, &msg, sizeof (msg));      
    }

  if (key == controls.fire)
    {
      msg.type = SD_MSG_FIRE;
      comWrite (&dgramChn, &msg, sizeof (msg));      
    }
}


/*
 * mouse
 *
 * Input:
 *    x
 *       Mouse cursor position (x).
 *    y
 *       Mouse cursor position (y).
 *
 * Description:
 *    GLUT callback for mouse events.
 *
 */

void
mouse (int     x,
       int     y)
{
  mouse_x = x;
  mouse_y = y;
}


/*
 * sendMouseCtrl
 *
 * Description:
 *    Send steering data to server.
 *
 */

void
sendMouseCtrl (void)
{
  float        distx;
  float        disty;
  message      msg;

  distx = ((float) mouse_x - ((float) win_width / 2)) /
    ((float) win_width / 2);
  disty = ((float) mouse_y - ((float) win_height / 2)) / 
    ((float) win_height / 2);

  if (distx > 1.0)
    {
      distx = 1.0;
    }
  if (disty > 1.0)
    {
      disty = 1.0;
    }
  if (distx < -1.0)
    {
      distx = -1.0;
    }
  if (disty < -1.0)
    {
      disty = -1.0;
    }

  msg.type = SD_MSG_PITCH;
  *((float *) &msg.data) = -disty;
  comWrite (&dgramChn, &msg, sizeof (msg));

  msg.type = SD_MSG_HEADING;
  *((float *) &msg.data) = -distx;
  comWrite (&dgramChn, &msg, sizeof (msg));
}


/*
 * mousePress
 *
 * Input:
 *    button
 *       Mouse button that changed state.
 *    state
 *       Indicates if button was pressed or released.
 *    x
 *       Cursor position (x).
 *    y
 *       Cursor position (y).
 *     
 * Description:
 *    GLUT callback for mouse button events.
 *
 */

void
mousePress (int    button,
	    int    state,
	    int    x,
	    int    y)
{
  switch (button)
    {
    case GLUT_LEFT_BUTTON:
      if (state == GLUT_DOWN)
	{
	  mouseCtrl = 1 - mouseCtrl;
	  if (mouseCtrl)
	    {
	      printf ("Mouse control.\n");
	      glutPassiveMotionFunc (mouse);
	    }
	  else
	    {
	      printf ("Keyboard control\n");
	      glutPassiveMotionFunc (NULL);
	    }
	}
      break;
    default:
      break;
    }
}


/*
 * init
 *
 * Description:
 *    Initialize client databases and set initial OpenGL state.
 *
 */

void
init (void)
{
  GLfloat     lightPos[4] = { 1.0, 0.0, 0.0, 0.0 };
  GLfloat     lightAmb[4] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat     lightDif[4] = { 1.0, 1.0, 0.9, 1.0 };
  GLfloat     lightSpc[4] = { 1.0, 1.0, 1.0, 1.0 };

  datagram[0].type = SD_EOD;
  initControls ();

  createResetDisplayList ();
  createRadarDL ();
  createStarfieldDL (displayOptions.wantTexturedStarfield);
  createBlastDL ();
  createVesselDL ();
  createExplosion ();
  createAsteroidDL ();
  
  glEnable (GL_LIGHT0);
  glLightfv (GL_LIGHT0, GL_POSITION, lightPos);
  glLightfv (GL_LIGHT0, GL_AMBIENT, lightAmb);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, lightDif);
  glLightfv (GL_LIGHT0, GL_SPECULAR, lightSpc);

  glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
}


/*
 * matchDOF
 *
 * Input:
 *    heading
 *       Direction of flight vector
 *    up
 *       Up vector
 *
 * Description:
 *    Creates a rotation matrix that maps the standard X, Y, Z
 *    coordinate system onto the right-hand coordinate system
 *    defined by the heading and up vectors. It also multiplies
 *    the current OpenGL transformation matrix with the rotation
 *    matrix.
 *
 */

void
matchDOF (vector    heading,
	  vector    up)
{
  GLfloat   m[16];
  vector    x;
  vector    y;
  vector    z;

  vecCopy (z, heading);
  vecNormalize (z ,z);
  vecCopy (y, up);
   
  /* X vector = Y cross Z */

  vecCross (x, y, z);

  /* Recompute Y = Z cross X */

  vecCross (y, z, x);

  vecNormalize (x, x);
  vecNormalize (y, y);

#define M(col,row)  m[col*4+row]
  M(0,0) = x[0];  M(0,1) = x[1];  M(0,2) = x[2];  M(0,3) = 0.0;
  M(1,0) = y[0];  M(1,1) = y[1];  M(1,2) = y[2];  M(1,3) = 0.0;
  M(2,0) = z[0];  M(2,1) = z[1];  M(2,2) = z[2];  M(2,3) = 0.0;
  M(3,0) = 0.0;   M(3,1) = 0.0;   M(3,2) = 0.0;   M(3,3) = 1.0;
#undef M

  glMultMatrixf (m);
}


/*
 * getWinDimensions
 *
 * Input:
 *    width
 *       Pointer to float variable where width is filled in.
 *    height
 *       Pointer to float variable where height is filled in.
 * 
 * Description:
 *    Fills in the current dimensions of the screen window in
 *    the specified variables.
 *
 */

void
getWinDimensions (float    *width,
		  float    *height)
{
  *width = (float) win_width;
  *height = (float) win_height;
}


/*
 * drawInfo
 *
 * Input:
 *    speed
 *       Speed in units/second.
 *    score
 *       Score.
 *    w
 *       Window width.
 *    h
 *       Window height.
 *
 * Description:
 *    Draw information text strings in the window.
 *
 */

void
drawInfo (float     speed,
	  int       score,
	  float     w,
	  float     h)
{
  int                i;
  char               str[80];
  float              l;

  glColor3f (1.0, 1.0, 0.0);

  glViewport (0.0, 0.0, w, h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluOrtho2D (0, w, 0, h);
  
  sprintf (str, "Speed: %0.2f", speed);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  l = (float) strlen (str) * 8.0;

  glRasterPos2f ((w / 2.0) - (l / 2.0),
		 0);

  for (i = 0; i < strlen (str); i++)
    {
      glutBitmapCharacter (GLUT_BITMAP_8_BY_13, str[i]);
    }

  sprintf (str, "Score: %1d", score);

  l = (float) strlen (str) * 8.0;

  glLoadIdentity ();
  glRasterPos2f ((w / 2.0) - (l / 2.0),
		 14.0);

  for (i = 0; i < strlen (str); i++)
    {
      glutBitmapCharacter (GLUT_BITMAP_8_BY_13, str[i]);
    }
}


/*
 * drawMouse
 *
 * Input:
 *    w, h
 *       Window dimensions.
 *    mx, my
 *       Mouse cursor position.
 *
 * Description:
 *    Draw a line to make mouse steering easier.
 *
 */

void
drawMouse (float    w,
	   float    h,
	   float    mx,
	   float    my)
{
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluOrtho2D (0.0, w, 0.0, h);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glDisable (GL_LIGHTING);
  glColor3f (1.0, 1.0, 1.0);
  glBegin (GL_LINES);
  glVertex2f (w / 2.0, h / 2.0);
  glVertex2f (mx, h - my);
  glEnd ();
}


/*
 * nearEnough
 *
 * Input:
 *    p
 *       Point.
 *    eye
 *       Pointer to eyepoint data.
 *
 * Description:
 *    Checks if a point is close enough to the camera to be drawn.
 *
 * Returns:
 *    true if point should be drawn.
 *    false otherwise.
 *
 */

int
nearEnough (vector p, eyePointXferData *eye)
{
  assert (eye);

  /* If point is beyond back clip plane, don't draw it */

  return (vecDist (p, eye->position) < SD_BACK_CLIP);
}


/*
 * pointVisible
 *
 * Input:
 *    p
 *       Point.
 *    eye
 *       Pointer to eyepoint data.
 *
 * Description:
 *    Checks if a point is visible to the camera.
 *
 * Returns:
 *    true if point should be drawn.
 *    false otherwise.
 *
 */

int
pointVisible (vector p, eyePointXferData *eye)
{
  vector v;

  assert (eye);

  /* Use same algorithm as back-face culling */

  vecSub (v, p, eye->position);
  return (vecDot (v, eye->heading) > 0);
}


/*
 * printTextMessage
 *
 * Input:
 *    msg
 *       Server message.
 *
 * Description:
 *    Extracts the text from a server text message and prints it to
 *    stdout.
 *
 */

void
printTextMessage (message   *msg)
{
  printf ("%s\n", msg->data);
}


/*
 * checkOptions
 *
 * Input:
 *    argc, argv
 *       Same as input to "main".
 *
 * Description:
 *    Checks for command line options and set internal state appropriately.
 *
 */

void
checkOptions (int     argc,
	      char   *argv[])
{
  int    i;
  float  factor;

  for (i = 0; i < argc; i++)
    {
      if (strcmp (argv[i], "-notextures") == 0)
	{
	  displayOptions.wantTextures = 0;
	}
      if (strcmp (argv[i], "-dithering") == 0)
	{
	  displayOptions.wantDithering = 1;
	}
      if (strcmp (argv[i], "-starfield") == 0)
	{
	  displayOptions.wantTexturedStarfield = 1;
	}
      if (strcmp (argv[i], "-2") == 0)
	{
	  win_width = 640;
	  win_height = 480;
	}
      if (strcmp (argv[i], "-ambient") == 0)
	{
	  factor = atof (argv[i + 1]);

	  if (factor >= 0.0 && factor <= 1.0)
	    {
	      displayOptions.globalAmbientFactor[0] = factor;
	      displayOptions.globalAmbientFactor[1] = factor;
	      displayOptions.globalAmbientFactor[2] = factor;	  
	    }
	}
    }
}


/*
 * initDisplayOptions
 *
 * Description:
 *    Initialize internal state.
 *
 */

void
initDisplayOptions (void)
{
  displayOptions.wantTextures = 1;
  displayOptions.wantAsteroidLighting = 1;
  displayOptions.wantDithering = 0;
  displayOptions.wantTexturedStarfield = 0;

  displayOptions.globalAmbientFactor[0] = 1.0;
  displayOptions.globalAmbientFactor[1] = 1.0;
  displayOptions.globalAmbientFactor[2] = 1.0;
}
