/***************************************************************************
 *
 * communication.h - Communication package for internet game applications
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************
 *
 *    The communication package simplifies socket communication between
 * internet game servers and clients by introducing an abstraction called
 * a channel. There are two types of channel, a stream channel and a
 * datagram channel. Once two channels are connected, the same function set
 * is used to read from and write to the channel.
 *    The connection procedure is somewhat different for the two types.
 * Stream sockets are connected through a server instance. The server
 * process creates a server instance and waits for clients to arrive.
 * Clients create a channel and connects to the server. The server can
 * be polled for client connections. If a client is found, a new channel
 * is returned and the connection is established.
 *    Datagram channels aren't really connected. Instead, the local port,
 * remote port and remote Internet address are bound to the channel at each
 * side of the communication link. There are different functions available
 * to handle this binding.
 *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include <communication.h>


/*
 * comCreateServer
 *
 * Input:
 *    server
 *       Pointer to a server.
 *    port
 *       Port. If 0, the system assigns a port to the server.
 *
 * Description:
 *    Prepares the server for client connection requests.
 *
 * Returns:
 *    Port the server is listening on if successful.
 *    -1 if unsuccessful. If the error is socket-related, errno is set.
 *
 */

int
comCreateServer (comServer    *server,
		 int           port)
{
  struct sockaddr_in   addr;

  assert (server != NULL);

  /* Create a socket */

  server->sockfd = socket (PF_INET, SOCK_STREAM, 0);
  if (server->sockfd < 0)
    {
      return -1;
    }

  /* Bind it to a port */

  memset ((char *) &addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl (INADDR_ANY);
  addr.sin_port = htons (port);
  if (bind (server->sockfd, (struct sockaddr *) &addr, sizeof (addr)))
    {
      return -1;
    }

  /* Listen for incoming clients */

  if (listen (server->sockfd, SOMAXCONN) < 0)
    {
      return -1;
    }

  /* Make server socket nonblocking */

  if (comSetNonblocking (server->sockfd) < 0)
    {
      return -1;
    }

  /* Return the assigned port of the server */

  server->port = comPort (server->sockfd);

  return server->port;
}


/*
 * comConnect
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *    host
 *       Name of a host.
 *    port
 *       Port of a server.
 *
 * Description:
 *    Tries to connect the channel to a server at the specified host,
 *    listening on the specified port.
 *
 * Returns:
 *    Local port of the channel if successful, -1 otherwise. Errno is set.
 *
 */

int
comConnect (comChannel    *channel, 
	    char          *host,
	    int            port)
{
  struct hostent       *hostp;
  struct sockaddr_in    addr;
  
  assert (channel != NULL);
  assert (host != NULL);

  /* Create a socket */

  channel->sockfd = socket (PF_INET, SOCK_STREAM, 0);
  if (channel->sockfd < 0)
    {
      return -1;
    }

  /* Get host internet address */

  hostp = gethostbyname (host);
  assert (hostp != NULL);        /* TODO: graceful exit */
  memset ((void *) &addr, 0, sizeof (struct sockaddr_in));
  addr.sin_family = PF_INET;
  memcpy ((void *) &addr.sin_addr, (void *)hostp->h_addr, hostp->h_length);
  addr.sin_port = htons (port);

  /* Connect */

  if (connect (channel->sockfd, 
	       (struct sockaddr *) &addr,
	       sizeof (struct sockaddr_in)) != 0)
    {
      return -1;
    }

  /* Return the assigned local port */

  channel->localPort = comPort (channel->sockfd);
  channel->remotePort = port;
  memcpy ((void *) &channel->remoteAddr, (void *) &addr, sizeof (addr));
  channel->type = COM_STREAM;

  return channel->localPort;
}


/*
 * comAccept
 *
 * Input:
 *    server
 *       Pointer to a server.
 *    channel
 *       Pointer to a channel.
 *
 * Description:
 *    Checks if there is a client connection request available on the
 *    server. If there is, accepts the client and initializes the channel.
 *
 * Returns:
 *    -1 if an error occurred. Errno is set.
 *    0 if no client connection requests was found.
 *    1 if successful.
 *
 */

int
comAccept (comServer     *server,
	   comChannel    *channel)
{
  int                 length;

  length = sizeof (struct sockaddr_in);
  errno = 0;
  channel->sockfd = accept (server->sockfd,
			    (struct sockaddr *) &channel->remoteAddr,
			    &length);

  if (channel->sockfd < 0 && errno == EWOULDBLOCK)
    {
      return 0;
    }

  if (channel->sockfd < 0)
    {
      return -1;
    }

  channel->type = COM_STREAM;
  channel->localPort = comPort (channel->sockfd);
  channel->remotePort = ntohs (channel->remoteAddr.sin_port);

  return 1;
}


/*
 * comBind
 *
 * Input:
 *    channel
 *       Pointer to a DATAGRAM channel.
 *    host
 *       Name of a host.
 *    remotePort
 *       Remote port.
 *    localPort
 *       Local port.
 *
 * Description:
 *    Tries to initialize a datagram channel.
 *
 * Returns:
 *    -1 if an error occurred. Errno is set.
 *    Local port if successful.
 *
 */

int
comBind (comChannel     *channel,
	 char           *host,
	 int             remotePort,
	 int             localPort)
{
  struct sockaddr_in     addr;
  struct hostent        *hostp;

  assert (channel != NULL);
  assert (host != NULL);

  /* Create a datagram socket */

  channel->sockfd = socket (PF_INET, SOCK_DGRAM, 0);
  if (channel->sockfd < 0)
    {
      return -1;
    }

  /* Initialize the remote address and port */

  hostp = gethostbyname (host);
  assert (hostp != NULL);        /* TODO: graceful exit */
  memset ((void *) &channel->remoteAddr, 0, sizeof (struct sockaddr_in));
  channel->remoteAddr.sin_family = PF_INET;
  memcpy ((void *) &(channel->remoteAddr.sin_addr), 
	  (void *) hostp->h_addr,
	  hostp->h_length);
  channel->remoteAddr.sin_port = htons (remotePort);

  /* Bind the local port */

  memset ((char *) &addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl (INADDR_ANY);
  addr.sin_port = htons (localPort);

  if (bind (channel->sockfd, (struct sockaddr *) &addr, sizeof (addr)) < 0)
    {
      return -1;
    }

  channel->localPort = comPort (channel->sockfd);
  channel->type = COM_DGRAM;
  channel->remotePort = remotePort;

  return channel->localPort;
}


/*
 * comBindAddr
 *
 * Input:
 *    channel
 *       Pointer to a DATAGRAM channel.
 *    addr
 *       Pointer to sockaddr_in structure containing remote address.
 *    localPort
 *       Local port.
 *
 * Description:
 *    Tries to initialize a datagram channel. Note that the port of
 *    the new channel will be the same as the one in the addr structure.
 *    Typically, it is changed later with comSetRemotePort.
 *
 * Returns:
 *    -1 if an error occurred. Errno is set.
 *    Local port if successful.
 *
 */

int
comBindAddr (comChannel          *channel,
	     struct sockaddr_in  *addr,
	     int                  localPort)
{
  struct sockaddr_in     tmpaddr;

  assert (channel != NULL);
  assert (addr != NULL);

  /* Create a datagram socket */

  channel->sockfd = socket (PF_INET, SOCK_DGRAM, 0);
  if (channel->sockfd < 0)
    {
      return -1;
    }

  /* Initialize the remote address and port */

  memcpy ((char *) &channel->remoteAddr,
	  (char *) addr,
	  sizeof (struct sockaddr_in));

  /* Bind the local port */

  memset ((char *) &tmpaddr, 0, sizeof (tmpaddr));
  tmpaddr.sin_family = AF_INET;
  tmpaddr.sin_addr.s_addr = htonl (INADDR_ANY);
  tmpaddr.sin_port = htons (localPort);

  if (bind (channel->sockfd,
	    (struct sockaddr *) &tmpaddr, 
	    sizeof (tmpaddr)) < 0)
    {
      return -1;
    }

  channel->localPort = comPort (channel->sockfd);
  channel->type = COM_DGRAM;
  channel->remotePort = ntohs(channel->remoteAddr.sin_port);

  return channel->localPort;
}


/*
 * comSetRemotePort
 *
 * Input:
 *    channel
 *       Pointer to communication channel.
 *    remotePort
 *       Remote port.
 *
 * Description:
 *    This function is used to complete a datagram channel initialized
 *    by comBindAddr. The reason this function exists is because there
 *    must be a way for a client and server to exchange remote datagram
 *    ports, but also to let the system assign these ports (by binding to
 *    port 0). Thus, the typical scenario is:
 *       1) A client connects to server. A sockaddr_in structure is aquired
 *          by the server.
 *       2) The server creates a datagram channel using comBindAddr.
 *       3) The server sends its local datagram port to the client.
 *       4) The client creates a datagram channel using comBind.
 *       5) The client sends its local datagram port to the server.
 *       6) The server sets the remote datagram port by calling
 *          comSetRemotePort.
 *
 */

void
comSetRemotePort (comChannel     *channel,
		  int             remotePort)
{
  assert (channel != NULL);

  channel->remoteAddr.sin_port = htons (remotePort);
}


/*
 * comClose
 *
 * Input:
 *   channel
 *      Pointer to a channel.
 *
 * Description:
 *   Closes the communication channel.
 *
 * Returns:
 *   0 if successful.
 *   -1 if unsuccessful. Errno is set.
 *
 */

int
comClose (comChannel     *channel)
{
  return close (channel->sockfd);
}


/*
 * comRead
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *    buffer
 *       Data buffer.
 *    nbytes
 *       Number of bytes to be read.
 *
 * Description:
 *    Reads a number of bytes into the specified buffer.
 *
 * Returns:
 *    Number of bytes read. -1 if an error occurred. Errno is set.
 *
 */

int
comRead (comChannel    *channel,
	 void          *buffer,
	 int            nbytes)
{
  struct sockaddr_in   addr;
  int                  addrlen;
#ifdef SD_SOLARIS
  int                  nread;
#endif

  assert (channel != NULL);
  assert (buffer != NULL);

  addrlen = sizeof (struct sockaddr_in);

  switch (channel->type)
    {
    case COM_STREAM:
#ifdef SD_SOLARIS
      nread = 0;
 
      while (nread != nbytes)
	{
	  nread = read (channel->sockfd, (char *) buffer, nbytes);
	  
	  if (nread != nbytes && errno != EAGAIN)
	    {
	      return -1;
	    }
	}

      return nread;
#else
      return read (channel->sockfd, (char *) buffer, nbytes);
#endif
      break;

    case COM_DGRAM:
      return recvfrom (channel->sockfd, 
		       (char *) buffer,
		       nbytes,
		       0, 
		       (struct sockaddr *) &addr,
		       &addrlen);
      break;

    default:
      fprintf (stderr, "comRead: Unkown channel type\n");
      /* TODO: Error here */
      break;
    }

  return -1;
}


/*
 * comWrite
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *    buffer
 *       Data buffer.
 *    nbytes
 *       Number of bytes to be written.
 *
 * Description:
 *    Writes a number of bytes to the channel from the specified buffer.
 *
 * Returns:
 *    Number of bytes written. -1 if an error occurred. Errno is set.
 *
 */

int comWrite (comChannel     *channel,
	      void           *buffer,
	      int             nbytes)
{
  assert (channel != NULL);
  assert (buffer != NULL);

  switch (channel->type)
    {
    case COM_STREAM:
      return write (channel->sockfd, (char *) buffer, nbytes);
      break;

    case COM_DGRAM:
      return sendto (channel->sockfd, 
		     (char *) buffer,
		     nbytes,
		     0, 
		     (struct sockaddr *) &channel->remoteAddr,
		     sizeof (struct sockaddr_in));
      break;

    default:
      fprintf (stderr, "comWrite: Unkown channel type\n");
      /* TODO: Error here */
      break;
    }

  return -1;
}


/*
 * comSetAddr
 *
 * Input:
 *    addr
 *       Pointer to a sockaddr_in structure to be filled in.
 *    host
 *       Host name in "127.0.0.0" format.
 *    port
 *       Port.
 *
 * Description:
 *    Fills in the sockaddr_in structure based on the data.
 *
 */

void
pxSetAddr (struct sockaddr_in *addr, char *host, int port)
{
  memset ((char *) addr, 0, sizeof (struct sockaddr_in));
  addr->sin_family = AF_INET;
  addr->sin_addr.s_addr = inet_addr (host);
  addr->sin_port = htons (port);
}


/*
 * comPort
 *
 * Input:
 *    sockfd
 *       Socket descriptor
 *
 * Description:
 *    Returns the port of the socket.
 *
 * Returns:
 *    Port of the socket if successful, -1 otherwise. Errno is set.
 *
 */

u_short
comPort (int sockfd)
{
  struct sockaddr_in   addr;
  int                  length;

  length = sizeof(struct sockaddr_in);
  if (getsockname(sockfd, (struct sockaddr *) &addr, &length) < 0)
    {
      return -1;
    }

  return ntohs(addr.sin_port);
}


/*
 * comSetNonblocking
 *
 * Input:
 *    sockfd
 *       Socket descriptor.
 *
 * Description:
 *    Sets the socket nonblocking flag.
 *
 * Returns:
 *    -1 if unsuccessful.
 *    0 if successful.
 *
 */

int
comSetNonblocking (int sockfd)
{
  if (fcntl (sockfd, F_SETFL, O_NONBLOCK) < 0)
    {
      return -1;
    }

  return 0;
}


/*
 * comPrintChannel
 *
 * Input:
 *    channel
 *       Pointer to channel.
 *
 * Description:
 *    Prints channel information to stdout.
 *
 */

void
comPrintChannel (comChannel   *channel)
{
  switch (channel->type)
    {
    case COM_STREAM:
      printf ("Type: Stream\n");
      break;
    case COM_DGRAM:
      printf ("Type: Datagram\n");
      break;
    default:
      printf ("Type: Unknown\n");
      break;
    }

  printf ("Socket: %d\n", channel->sockfd);
  printf ("Local port: %d\n", channel->localPort);
  printf ("Remote port: %d\n", channel->remotePort);
}


/*
 * comPoll
 *
 * Input:
 *    channel
 *       Pointer to a communication channel.
 *
 * Description:
 *    Polls the channel for available data to read.
 *
 * Returns:
 *    COM_POLL_NODATA if there is no data available to read.
 *    COM_POLL_DATA   if there is data available to read.
 *    COM_POLL_EXCP   if there is an exception on the channel.
 *    -1 if unsuccessful. Errno is set.
 *
 */

int
comPoll (comChannel    *channel)
{
#ifdef SD_SOLARIS
  struct pollfd    fds;

  fds.fd = channel->sockfd;
  fds.events = 0xFF;

  if (poll (&fds, 1, 0) == -1)
    {
      return -1;
    }

  if (fds.revents & POLLHUP)
    {
      return COM_POLL_EXCP;
    }

  if (fds.revents & POLLIN)
    {
      return COM_POLL_DATA;
    }

  return COM_POLL_NODATA;
#else
  fd_set            rfds;
  fd_set            efds;
  struct timeval    tv;

  FD_ZERO (&rfds);
  FD_ZERO (&efds);
  FD_SET (channel->sockfd, &rfds);
  FD_SET (channel->sockfd, &efds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  if (select (channel->sockfd + 1, &rfds, NULL, &efds, &tv) == -1)
    {
      return -1;
    }

  if (FD_ISSET (channel->sockfd, &efds))
    {
      return COM_POLL_EXCP;
    }

  if (FD_ISSET (channel->sockfd, &rfds))
    {
      return COM_POLL_DATA;
    }

  return COM_POLL_NODATA;
#endif
}
