/***************************************************************************
 *
 * linkedlist.c - Functions that implements a linked list ADT.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <stdlib.h>
#include <malloc.h>
#include <assert.h>
#include <stdio.h>

#include <linkedlist.h>


/*
 * llNewList
 *
 * Description:
 *    Creates a new linked list.
 *
 * Returns:
 *    Pointer to the new list.
 *
 */

linkedlist *
llNewList (void)
{
  linkedlist    *l;

  l = (linkedlist *) malloc (sizeof (linkedlist));
  assert (l);

  l->first = NULL;
  l->length = 0;
  l->cursor = NULL;
  l->last = NULL;

  return l;
}


/*
 * llFirst
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the first element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that list is
 *    empty).
 *
 */

int
llFirst (linkedlist *l)
{
#ifdef LL_DEBUG
  assert (l);
#endif

  l->cursor = l->first;

  if (l->cursor)
    {
      return 1;
    }
  else
    {
      return 0;
    }
}


/*
 * llNext
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the next element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that cursor was
 *    pointing to last element when the call was issued).
 *
 */

int
llNext (linkedlist *l)
{
#ifdef LL_DEBUG
  assert (l);
#endif

  if (l->cursor)
    {
      if ((l->cursor)->next)
	{
	  l->cursor = (l->cursor)->next;
	  return 1;
	}
    }

  return 0;
}


/*
 * llPrevious
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the previous element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that cursor was
 *    pointing to first element when the call was issued).
 *
 */

int
llPrevious (linkedlist *l)
{
#ifdef LL_DEBUG
  assert (l);
#endif

  if (l->cursor)
    {
      if ((l->cursor)->previous)
	{
	  l->cursor = (l->cursor)->previous;
	  return 1;
	}
    }  

  return 0;
}


/*
 * llLast
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Place the list cursor at the last element in the linked list.
 *
 * Returns:
 *    True if successful. False otherwise (indicates that list is empty).
 *
 */

int
llLast (linkedlist *l)
{
#ifdef LL_DEBUG
  assert (l);
#endif

  l->cursor = l->last;

  if (l->cursor)
    {
      return 1;
    }
  else
    {
      return 0;
    }
}


/*
 * llGetData
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Returns:
 *    Pointer to data contained in the list element pointed to by the
 *    list cursor.
 *
 */

void *
llGetData (linkedlist *l)
{
#ifdef LL_DEBUG
  assert (l);
#endif

  if (l->cursor)
    {
      return (l->cursor)->data;
    }
  else
    {
      return NULL;
    }
}


/*
 * llNewEntry
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *    data
 *       Pointer to data.
 *
 * Description:
 *    Creates a new list entry, containing the specified data. Inserts
 *    the new entry after the entry pointed to by the list cursor. The
 *    list cursor is then set to point to the new element.
 *
 * NOTE: The data is NOT copied into the element.
 *
 */

void
llNewEntry (linkedlist    *l, 
	    void          *data)
{
  linkedlistentry    *newEntry;
  linkedlistentry    *entry;

#ifdef LL_DEBUG
  assert (l);
#endif

  newEntry = (linkedlistentry *) malloc (sizeof (linkedlistentry));
  assert (newEntry);

  ++(l->length);
  newEntry->data = data;

  if (!l->cursor)
    {
      /* No elements in list */

      l->cursor = newEntry;
      l->first = newEntry;
      l->last = newEntry;
      
      newEntry->next = NULL;
      newEntry->previous = NULL;

      return;
    }

  entry = l->cursor;

  if (entry->next == NULL)
    {
      /* cursor is pointing to last element */

      l->cursor = newEntry;
      l->last = newEntry;

      newEntry->next = NULL;
      newEntry->previous = entry;
      
      entry->next = newEntry;

      return;
    }

  /* cursor is pointing to element with successor */

  l->cursor = newEntry;
  
  newEntry->next = entry->next;
  newEntry->previous = entry;

  (entry->next)->previous = newEntry;
  entry->next = newEntry;

  return;
}


/*
 * llDeleteEntry
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Description:
 *    Removes the element pointed to by the list cursor.
 *
 * Returns:
 *    Pointer to the data of the deleted element.
 *
 * NOTE: The element data is NOT deallocated.
 *
 */

void *
llDeleteEntry (linkedlist *l)
{
  linkedlistentry   *entry;
  void              *data;
  
#ifdef LL_DEBUG
  assert (l);
#endif

  if (!l->cursor)
    {
      /* No elements in list */

      return NULL;
    }

  entry = l->cursor;
  --(l->length);
  data = entry->data;

  if (entry->next == NULL && entry->previous == NULL)
    {
      /* One element in list only */

      l->cursor = NULL;
      l->first = NULL;
      l->last = NULL;
      free (entry);

      return data;
    }

  if (entry->previous == NULL)
    {
      /* Element is first in list */

      l->first = entry->next;
      l->cursor = entry->next;
      (entry->next)->previous = NULL;
      data = entry->data;
      free (entry);
      
      return data;
    }
  
  if (entry->next == NULL)
    {
      /* Element is last in list */

      l->last = entry->previous;
      l->cursor = l->first;
      (entry->previous)->next = NULL;
      data = entry->data;
      free (entry);

      return data;
    }

  /* Element is in between two other elements */

  l->cursor = l->first;
  (entry->previous)->next = entry->next;
  (entry->next)->previous = entry->previous;
  data = entry->data;
  free (entry);

  return data;
}


/*
 * llGetLength
 *
 * Input:
 *    l
 *       Pointer to a linked list.
 *
 * Returns:
 *    Number of elements in the list pointed to by l.
 *
 */

int
llGetLength (linkedlist *l)
{
#ifdef LL_DEBUG
  assert (l);
#endif

  return l->length;
}
