/***************************************************************************
 *
 * readppm.c - read ASCII .PPM files and store them as texture data.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>

#include <readppm.h>

#define MAX_LINE_LENGTH 65536



/*
 * readPPM
 *
 * Input:
 *    filename
 *       Name of ASCII .PPM file.
 *    func
 *       Pointer to function that creates texels from RGB data.
 *    w
 *       Pointer to integer where image width is stored.
 *    h
 *       Pointer to integer where image height is stored.
 *
 * Description: 
 *    Reads ASCII PPM data from the specified file. Creates a texture and
 *    stores the image dimensions in memory pointed to by w and h.
 *
 * Returns:
 *    Pointer to texture data if successful. NULL otherwise.
 *
 */

float *
readPPM (char              *filename,
	 createTexelFunc    func,
	 int               *w,
	 int               *h)
{
  FILE   *in;
  char    str[MAX_LINE_LENGTH];
  int     width, height;
  int     R, G, B;
  int     readMode;
  int     npixels;
  int     maxColorVal;
  float  *textureData;
  float  *ptr;

  assert (filename);
  
  in = fopen (filename, "r");
  if (!in)
    {
      return NULL;
    }

  readMode = 0;
  npixels = 0;

  while (!feof (in))
   {
     fscanf (in, "%s", str);
     if (str[0] == '#')         /* start of comment line */
       {
	 /* read until end of line */

	 while (getc (in) != '\n')
	   ;
       }
     else
       {
	 /* not a comment, so read whatever's appropriate */

	 switch (readMode)
	   {
	   case 0:
	     ++readMode;
	     break;
	   case 1:
	     sscanf (str, "%d", &width);
	     ++readMode;
	     break;
	   case 2:
	     sscanf (str, "%d", &height);
	     textureData = (float *) malloc (sizeof (float)
					     * width * height * 4);
	     if (!textureData)
	       {
		 fclose (in);
		 return NULL;
	       }
	     ptr = textureData;
	     ++readMode;
	     break;
	   case 3:
	     sscanf (str, "%d", &maxColorVal);
	     ++readMode;
	     break;
	   default:
	     switch (readMode)
	       {
	       case 4:
		 sscanf (str, "%d", &R);
		 ++readMode;
		 break;
	       case 5:
		 sscanf (str, "%d", &G);
		 ++readMode;
		 break;
	       case 6:
		 sscanf (str, "%d", &B);
		 (*func) (ptr, R, G, B, maxColorVal);
		 ptr += 4;
		 readMode = 4;
		 ++npixels;
		 break;
	       }
	     break;
	   }
       }
   }
  
  fclose (in);
  *w = width;
  *h = height;
  return textureData;
}


/*
 * readBinaryPPM
 *
 * Input:
 *    filename
 *       Name of binary/raw .PPM file.
 *    func
 *       Pointer to function that creates texels from RGB data.
 *    w
 *       Pointer to integer where image width is stored.
 *    h
 *       Pointer to integer where image height is stored.
 *
 * Description: 
 *    Reads binary PPM data from the specified file. Creates a texture and
 *    stores the image dimensions in memory pointed to by w and h.
 *
 * Returns:
 *    Pointer to texture data if successful. NULL otherwise.
 *
 */

float *
readBinaryPPM (char              *filename,
	       createTexelFunc    func,
	       int               *w,
	       int               *h)
{
  FILE   *in;
  char    str[MAX_LINE_LENGTH];
  int     width, height;
  char    R, G, B;
  int     readMode;
  int     npixels;
  int     maxColorVal;
  float  *textureData;
  float  *ptr;
  int     ready;
  int     x, y;

  assert (filename);
  
  in = fopen (filename, "r");
  if (!in)
    {
      return NULL;
    }

  readMode = 0;
  npixels = 0;
  ready = 0;

  while (!ready)
   {
     fscanf (in, "%s", str);
     if (str[0] == '#')         /* start of comment line */
       {
	 /* read until end of line */

	 while (getc (in) != '\n')
	   ;
       }
     else
       {
	 /* not a comment, so read whatever's appropriate */

	 switch (readMode)
	   {
	   case 0:
	     ++readMode;
	     break;
	   case 1:
	     sscanf (str, "%d", &width);
	     ++readMode;
	     break;
	   case 2:
	     sscanf (str, "%d", &height);
	     textureData = (float *) malloc (sizeof (float)
					     * width * height * 4);
	     if (!textureData)
	       {
		 fclose (in);
		 return NULL;
	       }
	     ptr = textureData;
	     ++readMode;
	     break;
	   case 3:
	     sscanf (str, "%d", &maxColorVal);
	     ++readMode;
	     ready = 1;
	     break;
	   }
       }
   }

  while (getc (in) != '\n')
    ;

  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
	{
	  R = getc (in);
	  G = getc (in);
	  B = getc (in);
	  (*func) (ptr, R, G, B, maxColorVal);
	  ptr += 4;
	}
    }
  
  fclose (in);
  *w = width;
  *h = height;
  return textureData;
}
