/***************************************************************************
 *
 * s_asteroid.c - Server side utility functions for asteroid entity.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <assert.h>
#include <stdio.h>

#include <vector.h>
#include <randomutils.h>

#include <s_world.h>
#include <s_asteroid.h>


/*
 * createAsteroids
 *
 * Description:
 *    Creates asteroid world entries.
 *
 */

void
createAsteroids (void)
{
  float             x;
  float             y;
  float             z;
  float             offset;
  int               i;
  boundingSphere    bs;
  asteroidData      data;

  x = (-SD_ASTEROIDS_X * SD_ASTEROID_BOX_SIZE) / 2;
  y = (-SD_ASTEROIDS_Y * SD_ASTEROID_BOX_SIZE) / 2;
  z = (-SD_ASTEROIDS_Z * SD_ASTEROID_BOX_SIZE) / 2;
  i = 0;

  while (x < (SD_ASTEROIDS_X * SD_ASTEROID_BOX_SIZE) / 2)
    {
      while (y < (SD_ASTEROIDS_Y * SD_ASTEROID_BOX_SIZE) / 2)
	{
	  while (z < (SD_ASTEROIDS_Z * SD_ASTEROID_BOX_SIZE) / 2)
	    {
	      ++i;
	      
	      data.position[0] = x + (SD_ASTEROID_BOX_SIZE / 2);
	      data.position[1] = y + (SD_ASTEROID_BOX_SIZE / 2);
	      data.position[2] = z + (SD_ASTEROID_BOX_SIZE / 2);

	      offset = randomBetweenMinusOneAndOne ();
	      offset *= (SD_ASTEROID_BOX_SIZE / 2) - SD_ASTEROID_RADIUS;
	      data.position[0] += offset;

	      offset = randomBetweenMinusOneAndOne ();
	      offset *= (SD_ASTEROID_BOX_SIZE / 2) - SD_ASTEROID_RADIUS;
	      data.position[1] += offset;

	      offset = randomBetweenMinusOneAndOne ();
	      offset *= (SD_ASTEROID_BOX_SIZE / 2) - SD_ASTEROID_RADIUS;
	      data.position[2] += offset;

	      data.angleX = 0.0;
	      data.angleY = 0.0;
	      data.angleZ = 0.0;
	      data.angleSpeedX = randomBetweenMinusOneAndOne () 
		* SD_ASTEROID_MAXSPEED;
	      data.angleSpeedY = randomBetweenMinusOneAndOne () 
		* SD_ASTEROID_MAXSPEED;
	      data.angleSpeedZ = randomBetweenMinusOneAndOne ()
		* SD_ASTEROID_MAXSPEED;
	      data.radius = SD_ASTEROID_RADIUS;

	      vecCopy (bs.position, data.position);
	      bs.radius = SD_ASTEROID_RADIUS;

	      newWorldEntry (SD_WE_ASTEROID,
			     asteroidUpdateFunc,
			     getAsteroidXferData,
			     asteroidCollisionFunc,
			     NULL,
			     &bs,
			     NULL,
			     (char *) &data,
			     sizeof (data));

	      z += SD_ASTEROID_BOX_SIZE;
	    }
	  z = (-SD_ASTEROIDS_Z * SD_ASTEROID_BOX_SIZE) / 2;
	  y += SD_ASTEROID_BOX_SIZE;
	}
      y = (-SD_ASTEROIDS_Y * SD_ASTEROID_BOX_SIZE) / 2;
      x += SD_ASTEROID_BOX_SIZE;
    }
}


/*
 * asteroidUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to (asteroid) world entry.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Runs asteroid simulation "timeslice" seconds.
 *
 */

void
asteroidUpdateFunc (worldEntry   *entry, 
		    double        timeslice)
{
  asteroidData      *asteroid;

  assert (entry->type == SD_WE_ASTEROID);

  asteroid = (asteroidData *) entry->data;

  asteroid->angleX += asteroid->angleSpeedX * timeslice;
  asteroid->angleY += asteroid->angleSpeedY * timeslice;
  asteroid->angleZ += asteroid->angleSpeedZ * timeslice;
}


/*
 * getAsteroidXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data to be filled in.
 *    entry
 *       Pointer to (asteroid) world entry.
 *
 * Description:
 *    Fills in transfer data for the asteroid pointed to by "entry".
 *
 */

void
getAsteroidXferData (xferData      *data,
		     worldEntry    *entry)
{
  asteroidXferData   xferData;
  asteroidData      *a;
  
  assert (entry != NULL);
  assert (data != NULL);
  assert (entry->type == SD_WE_ASTEROID);

  a = (asteroidData *) &entry->data;

  vecCopy (xferData.position, a->position);
  xferData.angleX = a->angleX;
  xferData.angleY = a->angleY;
  xferData.angleZ = a->angleZ;
  xferData.radius = a->radius;

  data->type = SD_ASTEROID;
  data->model = -1;
  memcpy ((char *) &data->data, (char *) &xferData, sizeof (xferData));
}


/*
 * asteroidCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (asteroid) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The asteroid entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void
asteroidCollisionFunc (worldEntry    *me,
		       worldEntry    *other)
{
  switch (other->type)
    {
    case SD_WE_BLAST:
      other->deleteMark = 1;
      break;

    case SD_WE_VESSEL:
      newExplosion (other);
      sendMsgDeath (other);
      initVessel ((vesselData *) other->data, &other->boundingSph);
      break;

    default:
      break;
    }

  return;
}


/*
 * outsideAsteroidField
 *
 * Input:
 *    p
 *       Point.
 *
 * Returns:
 *    true if p is outside the asteroid field, false otherwise.
 *
 */

int
outsideAsteroidField (vector p)
{
  return (p[0] < (-((float) SD_ASTEROIDS_X / 2) * SD_ASTEROID_BOX_SIZE)
	  || p[0] > (((float) SD_ASTEROIDS_X / 2) * SD_ASTEROID_BOX_SIZE)
	  || p[1] < (-((float) SD_ASTEROIDS_Y / 2) * SD_ASTEROID_BOX_SIZE)
	  || p[1] > (((float) SD_ASTEROIDS_Y / 2) * SD_ASTEROID_BOX_SIZE)
	  || p[2] < (-((float) SD_ASTEROIDS_Z / 2) * SD_ASTEROID_BOX_SIZE)
	  || p[2] > (((float) SD_ASTEROIDS_Z / 2) * SD_ASTEROID_BOX_SIZE));
}
