/***************************************************************************
 *
 * s_client.c - Server side abstraction of a connected client.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <assert.h>
#include <stdio.h>

#include <linkedlist.h>
#include <communication.h>

#include <general.h>
#include <transfer.h>
#include <s_client.h>
#include <s_simulation.h>


linkedlist             clients;        /* list containing clients */


/*
 * newClientEntry
 *
 * Input:
 *    stream
 *       Pointer to a stream communication channel.
 *    dgram
 *       Pointer to a datagram communication channel.
 *
 * Description:
 *    Adds a new entry to the client connection database.
 *
 * Returns:
 *    Pointer to the new entry if successful.
 *    NULL if unsuccessful. This indicates that the client
 *    connection database is full.
 *
 */

clientConnectionData *
newClientEntry (comChannel     *stream,
		comChannel     *dgram)
{
  clientConnectionData   *newClient;

  assert (stream != NULL);
  assert (dgram != NULL);
  
  newClient = (clientConnectionData *) malloc (sizeof (clientConnectionData));
  assert (newClient);

  memcpy ((char *) &(newClient->stream), 
	  (char *) stream,
	  sizeof (comChannel));
  memcpy ((char *) &(newClient->dgram),
	  (char *) dgram,
	  sizeof (comChannel));

  llNewEntry (&clients, newClient);

  return newClient;
}


/*
 * newClient
 *
 * Input:
 *    stream
 *       Pointer to stream communication channel.
 *    dgram
 *       Pointer to datagram communication channel.
 *
 * Description:
 *    Creates a new world entry that corresponds to a client.
 *
 */

void
newClient (comChannel    *stream,
	   comChannel    *dgram)
{
  vesselData              vessel;
  boundingSphere          bs;
  clientConnectionData   *conData;

  /* Add client to list of connected clients */

  conData = newClientEntry (stream, dgram);
  if (conData == NULL)
    {
      printf ("ERROR: Client connection database storage exhausted\n");
      exit (SD_ERROR);
    }

  /* Initialize client data */

  vessel.connectionData = conData;
  initVessel (&vessel, &bs);

  /* Add client to world database */

  conData->myData = newWorldEntry (SD_WE_VESSEL,
				   vesselUpdateFunc,
				   getVesselXferData,
				   vesselCollisionFunc,
				   NULL,
				   &bs,
				   NULL,
				   (char *) &vessel,
				   sizeof (vessel));

  if (conData->myData == NULL)
    {
      printf ("ERROR: World database storage exhausted\n");
      exit (SD_ERROR);
    }
}


/*
 * getClientList
 *
 * Returns:
 *    Pointer to the list of clients.
 *
 * NOTE:
 *    This function breaks the abstraction but makes it easier to
 *    iterate through the list of clients. A future version may
 *    return a list iterator instead.
 *
 */

linkedlist *
getClientList (void)
{
  return &clients;
}


/*
 * getNumberOfClientConnections
 *
 * Description:
 *    Returns the total number of client connections in the
 *    client connection database.
 *
 */

int
getNumberOfClientConnections (void)
{
  return llGetLength (&clients);
}


/*
 * deleteConnectionData
 *
 * Input:
 *    client
 *       Pointer to client connection data.
 *
 * Description:
 *    Removes the client connection data from the connection
 *    database.
 *
 */

void
deleteConnectionData (clientConnectionData   *client)
{
  int                     i;
  clientConnectionData   *c;

  assert (client != NULL);

  comClose (&client->stream);
  comClose (&client->dgram);

  llFirst (&clients);

  for (i = 0; i < llGetLength (&clients); i++)
    {
      c = (clientConnectionData *) llGetData (&clients);

      if (c == client)
	{
	  llDeleteEntry (&clients);
	  free (client);
	  return;
	}
      
      llNext (&clients);
    }  

  printf ("ERROR: Couldn't delete client connection data\n");
  exit (SD_ERROR);
}
