/***************************************************************************
 *
 * s_simulation.c - Server side simulation functions.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <vector.h>
#include <timeutils.h>

#include <general.h>
#include <s_simulation.h>
#include <s_world.h>
#include <s_asteroid.h>


/* -------- VESSELS (CLIENTS) -------- */

/*
 * initVessel
 *
 * Input:
 *    vessel
 *       Pointer to vessel data.
 *    bs
 *       Pointer to a bounding sphere.
 *
 * Description:
 *    Initializes a vessel data structure.
 *
 */

void
initVessel (vesselData      *vessel,
	    boundingSphere  *bs)
{  
  startPoint (vessel->position, vessel->heading, vessel->up);
  
  vessel->model = 0;
  vessel->animFrame = 0;

  vessel->speed = 0.0;
  vessel->activeBlasts = 0;
  vessel->lastFired = getTimestamp ();

  vessel->thrust = 0.0;
  vessel->headingThrust = 0.0;
  vessel->pitchThrust = 0.0;
  vessel->rollThrust = 0.0;

  vessel->lastFired = getTimestamp ();

  vessel->outsideAsteroidField = 0;
  vessel->warnings = 0;
  vessel->lastWarning = getTimestamp ();

  vecCopy (bs->position, vessel->position);
  bs->radius = SD_VESSEL_RADIUS;
}


/*
 * vesselUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to world database entry.
 *    timeslice
 *       Simulation timeslice.
 *
 * Description:
 *    Updates a vessel position and orientation based on its current
 *    state.
 *
 */

void
vesselUpdateFunc (worldEntry    *entry, 
		  double         timeslice)
{
  vesselData  *data;

  assert (entry != NULL);
  assert (entry->type == SD_WE_VESSEL);

  data = (vesselData *) entry->data;
  
  data->position[0] += (data->speed * timeslice) * data->heading[0];
  data->position[1] += (data->speed * timeslice) * data->heading[1];
  data->position[2] += (data->speed * timeslice) * data->heading[2];

  if (outsideAsteroidField (data->position))
    {
      if (data->outsideAsteroidField)
	{
	  /* We've been warned before */

	  if (timeDifference (data->lastWarning) > SD_VESSEL_WARNINGINTERVAL)
	    {
	      /* Time for a new warning */
	      
	      ++(data->warnings);
	      if (data->warnings == SD_VESSEL_WARNINGS)
		{
		  /* Maximum number of warnings reached, self-destruct */
		  
		  newExplosion (entry);
		  sendMsgDeath (entry);
		  initVessel (data, &entry->boundingSph);
		}
	      else
		{
		  /* Send new warning */

		  sendMsgWarning (entry);
		  data->lastWarning = getTimestamp ();
		}
	    }
	  else
	    {
	      /* Inside warning interval, so do nothing */
	    }
	}
      else
	{
	  /* We haven't been warned before */
	  
	  data->outsideAsteroidField = 1;
	  data->lastWarning = getTimestamp ();
	  data->warnings = 1;
	  sendMsgWarning (entry);
	}
    }
  else
    {
      /* We're inside the asteroid field */

      data->outsideAsteroidField = 0;
    }

  vecCopy ((entry->boundingSph).position, data->position);

  /* Update orientation */

  updateSpeed (data, timeslice);
  updateHeading (data, timeslice);
  updatePitch (data, timeslice);
  updateRoll (data, timeslice);
}


/*
 * updateSpeed
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel speed.
 *      
 */

void
updateSpeed (vesselData    *data,
	     double         timeslice)
{
  data->speed += data->thrust * SD_VESSEL_THRUSTSPEED * timeslice;

  if (data->speed > SD_VESSEL_MAXSPEED)
    {
      data->speed = SD_VESSEL_MAXSPEED;
    }

  if (data->speed < SD_VESSEL_MINSPEED)
    {
      data->speed = SD_VESSEL_MINSPEED;
    }
}


/*
 * updateHeading
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel heading.
 *      
 */

void
updateHeading (vesselData    *data,
	       double         timeslice)
{
  float        angle;

  angle = data->headingThrust 
    * timeslice
    * SD_VESSEL_ROTATIONSPEED
    * VEC_DEG_TO_RAD;

  vecRotate (data->heading, 
	     data->heading,
	     data->up,
	     angle);
}


/*
 * updatePitch
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel pitch.
 *      
 */

void
updatePitch (vesselData    *vessel,
	     double         timeslice)
{
  float        angle;
  vector   axis;

  angle = vessel->pitchThrust 
    * timeslice
    * SD_VESSEL_ROTATIONSPEED 
    * VEC_DEG_TO_RAD;

  vecCross (axis, vessel->heading, vessel->up);

  vecRotate (vessel->heading, 
	     vessel->heading,
	     axis, 
	     angle);
  vecRotate (vessel->up,
	     vessel->up,
	     axis,
	     angle);
}


/*
 * updateRoll
 *
 * Input:
 *    data
 *       Pointer to vessel data.
 *    timeslice
 *       Simulation timeslice in seconds.
 *
 * Description:
 *    Update vessel roll.
 *      
 */

void
updateRoll (vesselData    *vessel,
	    double         timeslice)
{
  float     angle;

  angle = vessel->rollThrust
    * timeslice
    * SD_VESSEL_ROTATIONSPEED
    * VEC_DEG_TO_RAD;

  vecRotate (vessel->up,
	     vessel->up,
	     vessel->heading,
	     angle);
}


/*
 * getVesselXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates vessel transfer data from world entry.
 *
 */

void
getVesselXferData (xferData          *data,
		   worldEntry        *entry)
{
  vesselXferData     xferData;
  vesselData        *c;

  assert (entry != NULL);
  assert (data != NULL);
  assert (entry->type == SD_WE_VESSEL);

  c = (vesselData *) &entry->data;

  vecCopy (xferData.position, c->position);
  vecCopy (xferData.heading, c->heading);
  vecCopy (xferData.up, c->up);
  xferData.animFrame = c->animFrame;
  xferData.speed = c->speed;

  data->type = SD_VESSEL;
  data->model = -1;
  memcpy ((char *) &data->data, (char *) &xferData, sizeof (xferData));
}


/*
 * vesselCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (vessel) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The vessel entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void
vesselCollisionFunc (worldEntry    *me,
		     worldEntry    *other)
{
  worldEntry    *parent;

  switch (other->type)
    {
    case SD_WE_VESSEL:
      newExplosion (me);
      newExplosion (other);
      sendMsgDeath (me);
      sendMsgDeath (other);
      initVessel ((vesselData *) me->data, &me->boundingSph);
      initVessel ((vesselData *) other->data, &other->boundingSph);
      break;

    case SD_WE_BLAST:
      newExplosion (me);
      sendMsgDeath (me);
      parent = findParent (other);
      sendMsgScore (parent);
      initVessel ((vesselData *) me->data, &me->boundingSph);
      other->deleteMark = 1;
      break;

    case SD_WE_ASTEROID:
      newExplosion (me);
      sendMsgDeath (me);
      initVessel ((vesselData *) me->data, &me->boundingSph);
      break;

    default:
      break;
    }
}


/*
 * getEyePointXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates eye point transfer data from world entry.
 *
 */

void
getEyePointXferData (xferData       *data,
		     worldEntry     *entry)
{
  eyePointXferData     xferData;
  vesselData          *c;

  assert (entry != NULL);
  assert (data != NULL);
  assert (entry->type == SD_WE_VESSEL);

  c = (vesselData *) &entry->data;

  vecCopy (xferData.position, c->position);
  vecCopy (xferData.heading, c->heading);
  vecCopy (xferData.up, c->up);
  xferData.speed = c->speed;

  data->type = SD_EYEPOINT;
  data->model = -1;
  memcpy ((char *) &data->data, (char *) &xferData, sizeof (xferData));
}


/*
 * sendMsgDeath
 *
 * Input:
 *    entry
 *       World entry.
 *
 * Description:
 *    Sends a message to a client that indicates that the client's vessel
 *    has been killed.
 *
 */

void
sendMsgDeath (worldEntry   *entry)
{
  vesselData    *v;
  comChannel    *stream;
  message        msg;

  assert (entry != NULL);
  assert (entry->type == SD_WE_VESSEL);

  v = (vesselData *) &entry->data;
  stream = &(v->connectionData->stream);
  msg.type = SD_MSG_DEATH;

  comWrite (stream, &msg, sizeof (msg));
}


/*
 * sendMsgScore
 *
 * Input:
 *    entry
 *       World entry.
 *
 * Description:
 *    Sends a message to a client that indicates that the client has
 *    scored a point.
 *
 */

void
sendMsgScore (worldEntry   *entry)
{
  vesselData    *v;
  comChannel    *stream;
  message        msg;

  assert (entry != NULL);
  assert (entry->type == SD_WE_VESSEL);

  v = (vesselData *) &entry->data;
  stream = &(v->connectionData->stream);
  msg.type = SD_MSG_SCORE;

  comWrite (stream, &msg, sizeof (msg));
}


/*
 * sendMsgText
 *
 * Input:
 *    entry
 *       World entry.
 *    str
 *       Text string.
 *
 * Description:
 *    Sends a text message to a client.
 *
 */

void
sendMsgText (worldEntry   *entry,
	     char         *str)
{
  vesselData    *v;
  comChannel    *stream;
  message        msg;
  int            len;

  assert (entry != NULL);
  assert (str != NULL);
  assert (entry->type == SD_WE_VESSEL);

  len = strlen (str);
  assert (len < SD_MESSAGEDATASIZE - 1);

  v = (vesselData *) &entry->data;
  stream = &(v->connectionData->stream);

  msg.type = SD_MSG_TEXT;
  memset (&msg.data, 0, SD_MESSAGEDATASIZE);
  memcpy ((char *) &msg.data, str, len);

  comWrite (stream, &msg, sizeof (msg));
}


/*
 * sendMsgWarning
 *
 * Input:
 *    entry
 *       World entry.
 *
 * Description:
 *    Sends a warning to a client. This is done when the client
 *    is venturing outside the asteroid field.
 *
 */

void
sendMsgWarning (worldEntry   *entry)
{
  vesselData    *v;
  comChannel    *stream;
  message        msg;
  
  assert (entry != NULL);
  assert (entry->type == SD_WE_VESSEL);
  
  v = (vesselData *) &entry->data;
  stream = &(v->connectionData->stream);
  msg.type = SD_MSG_WARNING;

  comWrite (stream, &msg, sizeof (msg));
}



/* -------- BLASTS -------- */


void
calculateBlastPositions (vesselData   *vessel,
			 vector        position1,
			 vector        position2)
{
  vector                     side;
  vector                     offset;

  /* Get direction orthogonal to heading and up vectors */

  vecCross (side, vessel->heading, vessel->up);
  vecNormalize (side, side);

  /*
   * Positions are 7.0 and -7.0 units along side vector
   * and SD_BLAST_RADIUS + 1.0 units in front of the vessel.
   *
   */

  vecMult (offset, 7.0, side);
  vecAdd (position1, vessel->position, offset);

  vecMult (offset, -7.0, side);
  vecAdd (position2, vessel->position, offset);

  vecMult (offset, SD_VESSEL_RADIUS + SD_BLAST_RADIUS + 1.0, vessel->heading);
  vecAdd (position1, position1, offset);
  vecAdd (position2, position2, offset);
}


/*
 * createBlast
 *
 * Input:
 *    entry
 *       Pointer to parent world entry.
 *    position
 *       Position of blast.
 *    heading
 *       Heading of blast.
 *    up
 *       Up direction of blast.
 *    speed
 *       Speed of parent.
 *
 * Description:
 *    Creates a new blast with the specified characteristics and
 *    adds it to the world database.
 *
 */

void
createBlast (worldEntry    *entry,
	     vector         position,
	     vector         heading,
	     vector         up,
	     float          speed)
{
  blastData                 blast;
  boundingSphere            bs;
  worldEntry               *b;

  vecCopy (blast.position, position);
  vecCopy (blast.heading, heading);
  vecCopy (blast.up, up);
  
  blast.animFrame = (char) 0;
  blast.speed = speed + SD_BLAST_SPEED;
  blast.fuel = SD_BLAST_FUEL;
  vecCopy (bs.position, blast.position);
  bs.radius = SD_BLAST_RADIUS;

  b = newWorldEntry (SD_WE_BLAST,
		     blastUpdateFunc,
		     getBlastXferData,
		     blastCollisionFunc,
		     blastDeregisterFunc,
		     &bs,
		     entry,
		     (char *) &blast,
		     sizeof (blast));

  if (b == NULL)
    {
      printf ("ERROR: World storage exhausted: couldn't create blast\n");
      exit (SD_ERROR);
    }

  b->parent = entry;  
}


/*
 * ctrlFire
 *
 * Input:
 *    vessel
 *       Pointer to vessel data.
 *
 * Description:
 *    Adds a blast to the world database.
 *
 */

void
ctrlFire (worldEntry    *entry,
	  char          *data)
{
  vesselData     *vessel;
  double          timediff;
  vector          pos1, pos2;

  vessel = (vesselData *) entry->data;

  if (vessel->activeBlasts >= SD_VESSEL_MAXBLASTS)
    {
      return;
    }

  timediff = timeDifference (vessel->lastFired);

  if (timediff < SD_VESSEL_RELOADTIME)
    {
      return;
    }
  
  vessel->lastFired = getTimestamp ();
  vessel->activeBlasts += 2;

  calculateBlastPositions (vessel, pos1, pos2);

  createBlast (entry, 
	       pos1,
	       vessel->heading, 
	       vessel->up,
	       vessel->speed);
  createBlast (entry, 
	       pos2,
	       vessel->heading, 
	       vessel->up,
	       vessel->speed);
}


/*
 * blastUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to world database entry.
 *    timeslice
 *       Simulation timeslice.
 *
 * Description:
 *    Updates a blast position and orientation based on its current
 *    state.
 *
 */

void
blastUpdateFunc (worldEntry    *entry, 
		 double         timeslice)
{
  blastData    *data;

  assert (entry != NULL);
  assert (entry->type == SD_WE_BLAST);

  data = (blastData *) entry->data;
  
  data->fuel -= 1;
  
  if (data->fuel <= 0.0)
    {
      entry->deleteMark = 1; 
    }
  else
    {
      data->fuel -= data->speed * timeslice;
      data->position[0] += (data->speed * timeslice) * data->heading[0];
      data->position[1] += (data->speed * timeslice) * data->heading[1];
      data->position[2] += (data->speed * timeslice) * data->heading[2];
      vecCopy ((entry->boundingSph).position, data->position);
    }
}


/*
 * getBlastXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates blast transfer data from world entry.
 *
 */

void
getBlastXferData (xferData          *data,
		  worldEntry        *entry)
{
  blastXferData     xferData;
  blastData        *b;

  assert (entry != NULL);
  assert (data != NULL);
  assert (entry->type == SD_WE_BLAST);

  b = (blastData *) &entry->data;

  vecCopy (xferData.position, b->position);
  vecCopy (xferData.heading, b->heading);
  vecCopy (xferData.up, b->up);
  xferData.animFrame = b->animFrame;

  data->type = SD_BLAST;
  data->model = -1;
  memcpy ((char *) &data->data, (char *) &xferData, sizeof (xferData));
}


/*
 * blastCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (blast) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The blast entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void
blastCollisionFunc (worldEntry    *me,
		    worldEntry    *other)
{
  return;
}


/*
 * blastDeregisterFunc
 *
 * Input:
 *    entry
 *       Pointer to (blast) world entry.
 *
 * Description:
 *    Callback for when a blast world entry is deleted.
 *
 */

void
blastDeregisterFunc (worldEntry   *entry)
{
  vesselData    *parent;

  assert (entry->type == SD_WE_BLAST);

  parent = (vesselData *) (entry->parent)->data;
  --(parent->activeBlasts);
}


/* -------- EXPLOSIONS -------- */


/*
 * explosionUpdateFunc
 *
 * Input:
 *    entry
 *       Pointer to world database entry.
 *    timeslice
 *       Simulation timeslice.
 *
 * Description:
 *    Updates an explosion based on its current state.
 *
 */

void
explosionUpdateFunc (worldEntry    *entry, 
		     double         timeslice)
{
  explosionData  *data;

  assert (entry != NULL);
  assert (entry->type == SD_WE_EXPLOSION);

  data = (explosionData *) entry->data;

  if (timeDifference (data->lastFrame) > SD_EXPLOSION_SPEED)
    {
      data->animFrame += (char) 1;
      if (data->animFrame >= (char) SD_EXPLOSION_FRAMES)
	{
	  entry->deleteMark = 1;
	}
    }
}


/*
 * getExplosionXferData
 *
 * Input:
 *    data
 *       Pointer to transfer data.
 *    entry
 *       Pointer to world entry.
 *
 * Description:
 *    Creates explosion transfer data from world entry.
 *
 */

void
getExplosionXferData (xferData          *data,
		      worldEntry        *entry)
{
  explosionXferData     xferData;
  explosionData        *e;

  assert (entry != NULL);
  assert (data != NULL);
  assert (entry->type == SD_WE_EXPLOSION);

  e = (explosionData *) &entry->data;

  vecCopy (xferData.position, e->position);
  xferData.animFrame = e->animFrame;

  data->type = SD_EXPLOSION;
  data->model = -1;
  memcpy ((char *) &data->data, (char *) &xferData, sizeof (xferData));
}


/*
 * explosionCollisionFunc
 *
 * Input:
 *    me
 *       Pointer to (explosion) world entry.
 *    other
 *       Pointer to a world entry.
 *
 * Description:
 *    The explosion entity pointed to by "me" has collided with the
 *    entity pointed to by "other". The function takes the appropriate
 *    action depending on the type of "other".
 *
 */

void
explosionCollisionFunc (worldEntry    *me,
			worldEntry    *other)
{
  return;
}


/*
 * newExplosion
 *
 * Input:
 *    entry
 *       Pointer to vessel.
 *
 * Description:
 *    Creates an explosion at the position of the specified vessel.
 *
 */

void
newExplosion (worldEntry    *entry)
{
  vesselData     *data;
  explosionData   e; 
  boundingSphere  bs;

  assert (entry != NULL);
  assert (entry->type == SD_WE_VESSEL);
  
  data = (vesselData *) &entry->data;

  vecCopy (e.position, data->position);
  e.animFrame = (char) 0;
  e.lastFrame = getTimestamp ();

  newWorldEntry (SD_WE_EXPLOSION,
		 explosionUpdateFunc,
		 getExplosionXferData,
		 explosionCollisionFunc,
		 NULL,
		 &bs,
		 entry,
		 (char *) &e,
		 sizeof (e));
}


