/***************************************************************************
 *
 * s_world.c - Server side world database interaction functions.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <math.h>

#include <randomutils.h>
#include <linkedlist.h>
#include <general.h>
#include <transfer.h>

#include <s_world.h>
#include <s_simulation.h>
#include <s_collision.h>
#include <s_asteroid.h>
#include <client.h>          /* for SD_BACK_CLIP */


linkedlist    *world;                       /* world database */


/*
 * newWorldEntry
 *
 * Input:
 *    type
 *       Type of new world database entry.
 *    uf
 *       Pointer to a world database entry update function. This is
 *       called during simulation.
 *    xf
 *       Pointer to a function that returns data suitable for
 *       transmission to a client.
 *    cf
 *       Pointer to function that responds to collisions.
 *    df
 *       Pointer to function that is called when object is deleted. May
 *       be NULL.
 *    bs
 *       Pointer to a bounding sphere.
 *    parent
 *       Pointer to parent. May be NULL.
 *    data
 *       Pointer to type-specific data.
 *    datasize
 *       Size of the type-specific data.
 *
 * Description:
 *    Adds a new entry to the world database.
 *
 * Returns:
 *    Pointer to the new world database entry if successful.
 *    NULL if unsuccsessful. This indicates that the world
 *    database is full.
 *
 */

worldEntry *
newWorldEntry (int                         type,
	       updateFunction              uf,
	       getXferDataFunction         xf,
	       collisionFunction           cf,
	       deregisterFunction          df,
	       boundingSphere             *bs,
	       worldEntry                 *parent,
	       char                       *data,
	       int                         datasize)
{
  worldEntry                              *newEntry;

  assert (uf != NULL);
  assert (xf != NULL);
  assert (cf != NULL);
  assert (data != NULL);
  assert (bs != NULL);
  assert (type >= 0 && type <= 3);   /* TODO: Remove this check */
  assert (datasize > 0 && datasize < SD_WORLDENTRYSIZE);

  newEntry = (worldEntry *) malloc (sizeof (worldEntry));
  assert (newEntry);

  newEntry->type = type;
  newEntry->updateFunc = uf;
  newEntry->getXferDataFunc = xf;
  newEntry->collisionFunc = cf;
  newEntry->deregisterFunc = df;
  newEntry->parent = parent;
  memcpy ((char *) &newEntry->boundingSph,
	  (char *) bs, 
	  sizeof (boundingSphere));
  memcpy ((char *) newEntry->data,
	  data,
	  datasize);

  llNewEntry (world, newEntry);

  return newEntry;
}


/*
 * deleteClient
 *
 * Input:
 *    client
 *       Pointer to client connection data.
 *
 * Description:
 *    Removes the client connection data from the list of
 *    clients. Also removes the client world entry from the
 *    world database.
 *
 */

void
deleteClient (clientConnectionData    *client)
{
  int           i;
  vesselData   *v;
  worldEntry   *entry;

  assert (client != NULL);
  
  llFirst (world);

  for (i = 0; i < llGetLength (world); i++)
    {
      entry = llGetData (world);
      
      if (entry->type == SD_WE_VESSEL)
	{
	  v = (vesselData *) entry->data;
	  if (v->connectionData == client)
	    {
	      /* Remove the client connection data */

	      deleteConnectionData (client);

	      /* Remove client's vessel from world database */

	      llDeleteEntry (world);
	      /* Call deregister function */
	      if (entry->deregisterFunc)
		{
		  (entry->deregisterFunc) (entry);
		}
	      free (entry);

	      printf ("Client disconnected.\n\n");
	      return;
	    }
	}
      
      llNext (world);
    }
}


/*
 * printWorld
 *
 * Description:
 *    Prints a representation of the world database on stdout.
 *    Legend: A = asteroid, B = blast, E = explosion, V = vessel.
 *
 */

void
printWorld (void)
{
  int           i;
  worldEntry   *entry;
  
  llFirst (world);

  for (i = 0; i < llGetLength (world); i++)
    {
      entry = llGetData (world);

      switch (entry->type)
	{
	case SD_WE_VESSEL:
	  printf ("V");
	  break;
	case SD_WE_ASTEROID:
	  printf ("A");
	  break;
	case SD_WE_BLAST:
	  printf ("B");
	  break;
	case SD_WE_EXPLOSION:
	  printf ("E");
	  break;
	}
      
      llNext (world);
    }

  printf ("\n");
}


/*
 * deleteMarkedWorldEntries
 *
 * Description:
 *    Steps through the world database and removes all world entries that
 *    have the "deleteMark" flag set.
 *
 */

void
deleteMarkedWorldEntries (void)
{
  worldEntry    *entry;

  llFirst (world);

  do
  {
    entry = llGetData (world);
    if (entry)
      {
	if (entry->deleteMark)
	  {
	    llDeleteEntry (world);
	    /* Call deregister function */
	    if (entry->deregisterFunc)
	      {
		(entry->deregisterFunc) (entry);
	      }
	    free (entry);
	    break;
	  }
      }
  }
  while (llNext (world));
}


/*
 * updateWorld
 *
 * Input:
 *    timeslice
 *       How long the call to updateWorld should run the simulation
 *       in seconds. I.e. when the call to updateWorld is finished,
 *       the simulation should have progressed 'timeslice' seconds.
 *
 * Description:
 *    Updates the world database by calling update functions.
 *
 */

void
updateWorld (double    timeslice)
{
  worldEntry           *entry;
  int                   i;
  
  llFirst (world);
  
  for (i = 0; i < llGetLength (world); i++)
    {
      entry = llGetData (world);
      entry->deleteMark = 0;
      
      (entry->updateFunc) (entry, timeslice);
      llNext (world);
    }
}


/*
 * clearWorld
 *
 * Description:
 *    Initializes the world database.
 *
 */

void
clearWorld (void)
{
  world = llNewList ();
}


/*
 * fetchVesselData
 *
 * Input:
 *    client
 *       Pointer to client connection data.
 *
 * Description:
 *    Returns a pointer to the vessel data structure that
 *    corresponds to the specified client connection data.
 *
 * Returns:
 *    Pointer to vessel data structure if successful.
 *    NULL otherwise.
 *
 */

vesselData *
fetchVesselData (clientConnectionData     *client)
{
  return (vesselData *) ((client->myData)->data);
}


/*
 * entryVisible
 *
 * Returns: true if "entry" is visible to "client".
 *
 */

int
entryVisible (worldEntry   *entry,
	      worldEntry   *client)
{
  vesselData    *clientVessel;
  asteroidData  *asteroid;
  blastData     *blast;
  explosionData *expl;
  vesselData    *vessel;

  clientVessel = (vesselData *) client->data;

  switch (entry->type)
    {
    case SD_WE_ASTEROID:
      asteroid = (asteroidData *) entry->data;
      if (vecDist (asteroid->position,
		   clientVessel->position) > SD_BACK_CLIP)
	{
	  return SD_FALSE;
	}
      break;

    case SD_WE_BLAST:
      blast = (blastData *) entry->data;
      if (vecDist (blast->position,
		   clientVessel->position) > SD_BACK_CLIP)
	{
	  return SD_FALSE;
	}
      break;

    case SD_WE_EXPLOSION:
      expl = (explosionData *) entry->data;
      if (vecDist (expl->position,
		   clientVessel->position) > SD_BACK_CLIP)
	{
	  return SD_FALSE;
	}
      break;

    case SD_WE_VESSEL:
      vessel = (vesselData *) entry->data;
      if (vecDist (vessel->position,
		   clientVessel->position) > SD_BACK_CLIP)
	{
	  return SD_FALSE;
	}
      break;
    }

  return SD_TRUE;
}


/*
 * sendWorld
 *
 * Description:
 *    Send world to all clients.
 *
 */

void
sendWorld (void)
{
  int                     j;
  xferData                datagram[SD_DATALENGTH];
  linkedlist             *l;
  clientConnectionData   *conData;
  clientConnectionData   *client;
  worldEntry             *entry;
  worldEntry             *clientWE;


  l = getClientList ();
  if (llGetLength (l) <= 0)
    {
      return;
    }

  llFirst (l);
  do
    {
      client = (clientConnectionData *) llGetData (l);

      clientWE = client->myData;
      j = 0;
      llFirst (world);
      
      do
	{
	  /*
	   * Get transfer data from world entry object:
	   *
	   * If the object is a client, we need to check if it is
	   * the client we're sending the datagram to. If it is,
	   * An eye point should be sent, not a vessel.
	   *
	   */
	  
	  entry = llGetData (world);

	  if (entry->type == SD_WE_VESSEL)
	    {
	      conData = ((vesselData *) &(entry->data))->connectionData;
	      
	      if (client == conData)
		{
		  getEyePointXferData (&datagram[j], entry);
		  ++j;
		}
	      else
		{
		  if (entryVisible (entry, clientWE))
		    {
		      (entry->getXferDataFunc) (&datagram[j], entry);
		      ++j;
		    }
		}
	    }
	  else
	    {
	      if (entryVisible (entry, clientWE))
		{
		  (entry->getXferDataFunc) (&datagram[j], entry);
		  ++j;
		}
	    }
	}
      while (llNext (world));
      
      datagram[j].type = SD_EOD;
      comWrite (&(client->dgram), 
		(char *) &datagram[0],
		SD_DATALENGTH * sizeof (xferData));
    }
  while (llNext (l));
}


/*
 * startPoint
 *
 * Input:
 *    position
 *       Vector to be filled in with start position.
 *    heading
 *       Vector to be filled in with start heading.
 *    up
 *       Vector to be filled in with up direction.
 *
 * Description:
 *    Calculates a new vessel/client start position based on
 *    the current entities in the database.
 *
 */

void
startPoint (vector position,
	    vector heading,
	    vector up)
{
  int            a;
  int            i;
  asteroidData  *asteroid;
  worldEntry    *entry;
  vector         dir;
  vector         avgHeading;
  int            nvessels;
  vesselData    *vessel;
  
  /* Select an asteroid */

  a = (int) floor ((randomBetweenZeroAndOne () * (SD_ASTEROIDS - 1)) + 1.0);

  /* Fetch the asteroid data */

  i = 0;
  avgHeading[0] = avgHeading[1] = avgHeading[2] = 0.0;
  nvessels = 0;

  llFirst (world);
  do
    {
      entry = llGetData (world);
      if (entry->type == SD_WE_ASTEROID)
	{
	  ++i;
	  if (i == a)
	    {
	      asteroid = (asteroidData *) entry->data;
	    }
	}
      if (entry->type == SD_WE_VESSEL)
	{
	  vessel = (vesselData *) entry->data;
	  ++nvessels;
	  vecAdd (avgHeading, avgHeading, vessel->heading);
	}
    }
  while (llNext (world));

  /* Create a random direction vector */

  dir[0] = randomBetweenMinusOneAndOne ();
  dir[1] = randomBetweenMinusOneAndOne ();
  dir[2] = randomBetweenMinusOneAndOne ();
  vecNormalize (dir, dir);

  /* Multiply with asteroid radius + vessel radius + offset */

  vecMult (dir, asteroid->radius + 2 * SD_VESSEL_RADIUS + 1.0, dir);

  /* Set position to asteroid position + direction vector */

  vecAdd (position, dir, asteroid->position);
  
  /* Set heading to average heading vector */

  if (nvessels > 0)
    {
      vecMult (avgHeading, 1 / nvessels, avgHeading);
      vecNormalize (heading, avgHeading);
    }
  else
    {
      heading[0] = 0.0;
      heading[1] = 0.0;
      heading[2] = 1.0;
    }

  /* Set up vector to cross product between heading and direction vector */

  vecCross (up, heading, dir);
  vecNormalize (up, up);
}


/*
 * findParent
 *
 * Input:
 *    entry
 *       Pointer to world entry.
 *
 * Returns:
 *    Parent of the world entry. May be NULL.
 *
 */

worldEntry *
findParent (worldEntry  *entry)
{
  return entry->parent;
}


/*
 * checkCollisions
 *
 * Description:
 *    Checks for collisions among the objects of the world.
 *
 */

void
checkCollisions (void)
{
  worldEntry           *entryI;
  worldEntry           *entryJ;
  linkedlistentry      *elemI;    /* TODO: Breaking the abstraction!!!! */
  linkedlistentry      *elemJ;

  elemI = world->first;
  elemJ = world->first;

  while (elemI)
    {
      entryI = elemI->data;
      
      if (!entryI->deleteMark)
	{
	  while (elemJ)
	    {
	      entryJ = elemJ->data;
	      
	      if (entryI != entryJ && !entryJ->deleteMark)
		{
		  if (collisionSpheres (&entryI->boundingSph,
					&entryJ->boundingSph))
		    {
		      entryJ->deleteMark = 0;
		      (entryI->collisionFunc) (entryI, entryJ);
		    }	      
		}
	      
	      elemJ = elemJ->next;
	    }
	}

      elemJ = elemI->next;
      elemI = elemI->next;
    }

  deleteMarkedWorldEntries ();
}
