/***************************************************************************
 *
 * server.c - SpaceDuel3D server.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>

#include <communication.h>
#include <timeutils.h>
#include <randomutils.h>
#include <linkedlist.h>

#include <server.h>
#include <general.h>
#include <transfer.h>
#include <s_simulation.h>
#include <s_world.h>
#include <s_client.h>
#include <s_userinput.h>
#include <s_asteroid.h>


void lookForClient (comServer *server);
void handshake (comChannel *channel);
void createDatagramChannel (comChannel *stream, comChannel *dgram);



int
main (int     argc,
      char   *argv[])
{
  int         serverport;
  comServer   server;

  timestamp   start;
  double      elapsed;      /* elapsed time in milliseconds */
  double      timeslice;    /* timeslice in seconds */


  printf ("================================================\n");
  printf ("SpaceDuel3D Server\n\n");
  printf ("Version 1.0 BETA\n");
  printf ("Copyright (C) 1998 Gustav Taxen.\n");
  printf ("------------------------------------------------\n");
  printf ("SpaceDuel3D comes with ABSOLUTELY NO WARRANTY.\n");
  printf ("This is free software, and you are welcome\n");
  printf ("to redistribute it under certain conditions.\n");
  printf ("Please read the file 'COPYING' in the root\n");
  printf ("directory of the SpaceDuel3D distribution.\n");
  printf ("================================================\n\n");

  /* Check command line arguments */

  if (argc < 2)
    {
      printf ("Usage: %s port\n", argv[0]);
      return SD_ERROR;
    }

  /* Create server instance */

  serverport = atoi (argv[1]);
  serverport = comCreateServer (&server, serverport);
  
  if (serverport < 0)
    {
      printf ("ERROR: %s\n", strerror (errno));
      return SD_ERROR;
    }

  printf ("Listening for clients on port %d.\n\n", serverport);

  /* Clear world database */

  randomize ();  
  clearWorld ();
  createAsteroids ();

  /* MAIN LOOP */

  timeslice = SD_PACKET_FREQUENCY * TIME_MSEC_TO_SEC;

  for (;;)
    {
      start = getTimestamp ();
      
      lookForClient (&server);
      updateWorld (timeslice);
      checkCollisions ();
      sendWorld ();
      readInput ();

      elapsed = timeDifference (start) * TIME_SEC_TO_MSEC;
      timeSleepMsec ((int) ((double) SD_PACKET_FREQUENCY - elapsed));
    }

  return SD_OK;
}


/*
 * lookForClient
 *
 * Input:
 *    server
 *       Pointer to a server object
 *
 * Description:
 *    Polls the server object for clients.
 *
 */

void
lookForClient (comServer *server)
{
  comChannel   stream;
  comChannel   dgram;
  int          res;

  res = comAccept (server, &stream);

  if (res < 0)
    {
      printf ("ERROR: %s\n", strerror (errno));
      exit (SD_ERROR);
    }

  if (res > 0)
    {
      printf ("New client connection request:\n");
      handshake (&stream);
      createDatagramChannel (&stream, &dgram);
      newClient (&stream, &dgram);      
      printf ("New client accepted.\n\n");
    }
}


/*
 * createDatagramChannel
 *
 * Input:
 *    stream
 *       Pointer to stream channel.
 *    dgram
 *       Pointer to datagram channel.
 *
 * Description:
 *    Creates a datagram channel and exchanges port numbers with
 *    the client through the stream channel.
 *
 */

void
createDatagramChannel (comChannel     *stream,
		       comChannel     *dgram)
{
  int    localPort;
  int    remotePort;

  localPort = comBindAddr (dgram, &(stream->remoteAddr), 0);

  printf ("Using port %d for datagram communication.\n", localPort);
  comWrite (stream, &localPort, sizeof (localPort));

  comRead (stream, &remotePort, sizeof (remotePort));
  printf ("Remote datagram port is %d.\n", remotePort);
  comSetRemotePort (dgram, remotePort);
}


/*
 * handshake
 *
 * Input:
 *    channel
 *       Pointer to a channel.
 *
 * Description:
 *    Shakes hands with a client through a channel. If unsuccessful,
 *    the connection is closed.
 *
 */

void
handshake (comChannel *channel)
{
  if (readHandshake (channel) == SD_ERROR)
    {
      comClose (channel);
      printf ("WARNING: Handshaking failed, refusing client.\n");
      return;
    }
  
  if (sendHandshake (channel) == SD_ERROR)
    {
      comClose (channel);
      printf ("WARNING: Handshaking failed, refusing client.\n");
      return;
    }
}


