/***************************************************************************
 *
 * vector.c - Vector math utilities.
 *
 ***************************************************************************
 *
 * SpaceDuel3D, version 1.0.
 * Copyright (C) 1998 Gustav Taxen.
 * This is free software with ABSOLUTELY NO WARRANTY.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <stdio.h>
#include <math.h>
#include <vector.h>


/*
 * vecPrint
 *
 * Input:
 *    v
 *       Vector.
 *
 * Description:
 *    Prints the vector on stdout.
 *
 */

void
vecPrint (vector v)
{
  printf ("[%f %f %f]\n", v[0], v[1], v[2]);
}


/*
 * vecCopy
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *
 * Description:
 *    Copies the contents of the source vector into the destination vector.
 *
 */

void
vecCopy (vector dest, vector src)
{
  dest[0] = src[0];
  dest[1] = src[1];
  dest[2] = src[2];
}


/*
 * vecLength
 *
 * Input:
 *    u
 *       Vector.
 *
 * Returns:
 *    Norm of the vector.
 *
 */

float
vecLength (vector u)
{
  return sqrt (u[0] * u[0] + u[1] * u[1] + u[2] * u[2]);
}


/*
 * vecLengthSq
 *
 * Input:
 *    u
 *       Vector
 *
 * Returns:
 *    Square of the norm of the vector.
 *
 */

float
vecLengthSq (vector u)
{
  return (u[0] * u[0] + u[1] * u[1] + u[2] * u[2]);
}


/*
 * vecDist
 *
 * Input:
 *    u
 *       Point.
 *    v
 *       Point.
 *
 * Returns:
 *    Geometrical distance between points u and v.
 *
 */

float
vecDist (vector   u,
	 vector   v)
{
  return sqrt ((v[0] - u[0]) * (v[0] - u[0]) +
	       (v[1] - u[1]) * (v[1] - u[1]) +
	       (v[2] - u[2]) * (v[2] - u[2]));
}


/*
 * vecDistSq
 *
 * Input:
 *    u
 *       Point.
 *    v
 *       Point.
 *
 * Returns:
 *    Square of the geometrical distance between points u and v.
 *
 */

float
vecDistSq (vector   u,
	   vector   v)
{
  return ((v[0] - u[0]) * (v[0] - u[0]) +
	  (v[1] - u[1]) * (v[1] - u[1]) +
	  (v[2] - u[2]) * (v[2] - u[2]));
}


/*
 * vecSub
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts (u - v) in "dest".
 * 
 */

void
vecSub (vector     dest, 
	vector     u,
	vector     v)
{
  vector    res;

  res[0] = u[0] - v[0];
  res[1] = u[1] - v[1];
  res[2] = u[2] - v[2];

  vecCopy (dest, res);
}


/*
 * vecAdd
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts (u + v) in "dest".
 * 
 */

void
vecAdd (vector     dest, 
	vector     u,
	vector     v)
{
  vector     res;

  res[0] = u[0] + v[0];
  res[1] = u[1] + v[1];
  res[2] = u[2] + v[2];

  vecCopy (dest, res);
}


/*
 * vecMult
 *
 * Input:
 *    dest
 *       Destination vector.
 *    c
 *       Scalar constant.
 *    u
 *       Vector.
 *
 * Description:
 *    Puts c * u in "dest".
 * 
 */

void
vecMult (vector     dest, 
	 float      c,
	 vector     u)
{
  vector    res;

  res[0] = u[0] * c;
  res[1] = u[1] * c;
  res[2] = u[2] * c;

  vecCopy (dest, res);
}


/*
 * vecDot
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts dot product of u and v in "dest".
 * 
 */

float
vecDot (vector    u,
	vector    v)
{
  return (u[0] * v[0] + u[1] * v[1] + u[2] * v[2]);
}


/*
 * vecAngle
 *
 * Input:
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Returns:
 *    Geometrical angle between u and v in radians.
 * 
 */

float
vecAngle (vector   u,
	  vector   v)
{
  return acos (vecDot (u, v) / (vecLength (u) * vecLength (v)));
}


/*
 * vecCross
 *
 * Input:
 *    dest
 *       Destination vector.
 *    u
 *       Vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts cross product of u and v in "dest".
 * 
 */

void
vecCross (vector    dest,
	  vector    u,
	  vector    v)
{
  vector res;

  res[0] = u[1] * v[2] - u[2] * v[1];
  res[1] = u[2] * v[0] - u[0] * v[2];
  res[2] = u[0] * v[1] - u[1] * v[0];

  vecCopy (dest, res);
}


/*
 * vecNormalize
 *
 * Input:
 *    dest
 *       Destination vector.
 *    v
 *       Vector.
 *
 * Description:
 *    Puts (v / |v|) in "dest".
 * 
 */

void
vecNormalize (vector dest, vector v)
{
  float   length;

  length = vecLength (v);
  
  if (length != 0)
    {
      vecMult (dest, 1 / length, v);
    }
}


/*
 * vecDistPointToPlane
 *
 * Input:
 *    r
 *       Point.
 *    p
 *       Plane described by plane equation.
 *
 * Returns:
 *    Geometrical distance (signed) between point and plane.
 *
 */

float
vecDistPointToPlane (vector    r,
		     plane     p)
{
  return (p[0] * r[0] + p[1] * r[1] + p[2] * r[2] + p[3]);
}


/*
 * vecPlaneFromPointAndNormal
 *
 * Input:
 *    dest
 *       Destination plane.
 *    p
 *       Point on a plane.
 *    N
 *       Normal of a plane.
 *
 * Description:
 *    Calculates the plane equation corresponding to the plane
 *    defined by p and N.
 *
 */

void
vecPlaneFromPointAndNormal (plane    dest,
			    vector   p,
			    vector   N)
{
  dest[0] = N[0];
  dest[1] = N[1];
  dest[2] = N[2];
  dest[3] = -((N[0] * p[0]) + (N[1] * p[1]) + (N[2] * p[2]));
}


/*
 * vecSpherical
 *
 * Input:
 *    dest
 *       Destination point.
 *    theta
 *       Angle in radians.
 *    phi
 *       Angle in radians.
 *
 * Description:
 *    Calculates the vector [x y z] that corresponds to the given
 *    theta and phi in the spherical coordinates transformation
 *
 *    x = sin (theta) * cos (phi)
 *    y = sin (theta) * sin (phi)
 *    z = cos (theta)
 *
 *    where
 *
 *    0 <= theta <= PI
 *    0 <= phi <= 2*PI
 *
 */

void
vecSpherical (vector dest, float theta, float phi)
{
  dest[0] = sin (theta) * cos (phi);
  dest[1] = sin (theta) * sin (phi);
  dest[2] = cos (theta);
}


/*
 * vecRotateX
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates the source vector around the X axis.
 *
 */

void
vecRotateX (vector     dest,
	    vector     src,
	    float      angle)
{
  float    s;
  float    c;
  vector   result;

  s = sin (angle * VEC_DEG_TO_RAD);
  c = cos (angle * VEC_DEG_TO_RAD);

  result[0] = src[0];
  result[1] = src[1] * c + src[2] * s;
  result[2] = -src[1] * s + src[2] * c;

  vecCopy (dest, result);
}


/*
 * vecRotateY
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates the source vector around the Y axis.
 *
 */

void
vecRotateY (vector      dest,
	    vector      src,
	    float       angle)
{
  float    s;
  float    c;
  vector   result;

  s = sin (angle * VEC_DEG_TO_RAD);
  c = cos (angle * VEC_DEG_TO_RAD);

  result[0] = src[0] * c - src[2] * s;
  result[1] = src[1];
  result[2] = src[0] * s + src[2] * c;

  vecCopy (dest, result);
}


/*
 * vecRotateZ
 *
 * Input:
 *    dest
 *       Destination vector.
 *    src
 *       Source vector.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates the source vector around the Z axis.
 *
 */

void
vecRotateZ (vector     dest,
	    vector     src,
	    float      angle)
{
  float   s;
  float   c;
  vector  result;

  s = sin (angle * VEC_DEG_TO_RAD);
  c = cos (angle * VEC_DEG_TO_RAD);

  result[0] = src[0] * c + src[1] * s;
  result[1] = -src[0] * s + src[1] * c;
  result[2] = src[2];
  
  vecCopy (dest, result);
}


/*
 * vecRotate
 *
 * Input:
 *    dest
 *       Destination point.
 *    p
 *       Source point.
 *    axis
 *       Normalized direction vector that defines an axis of rotation.
 *    angle
 *       Angle in radians.
 *
 * Description:
 *    Rotates a point around an arbitrary axis.
 *
 */

void
vecRotate (vector dest, vector p, vector axis, float angle)
{
  float   c;
  float   s;
  float   t;
  float   x, y, z;
  float   x2, y2, z2;
  float   R[3][3];
  vector  result;

  c = cos (angle);
  s = sin (angle);
  t = 1 - c;
  x = axis[0];
  y = axis[1];
  z = axis[2];
  x2 = x * x;
  y2 = y * y;
  z2 = z * z;

  R[0][0] = (t * x2) + c;
  R[1][0] = (t * x * y) - (s * z);
  R[2][0] = (t * x * z) + (s * y);
  
  R[0][1] = (t * x * y) + (s * z);
  R[1][1] = (t * y2) + c;
  R[2][1] = (t * y * z) - (s * x);

  R[0][2] = (t * x * z) - (s * y);
  R[1][2] = (t * y * z) + (s * x);
  R[2][2] = (t * z2) + c;

  result[0] = (p[0] * R[0][0]) + (p[1] * R[1][0]) + (p[2] * R[2][0]);
  result[1] = (p[0] * R[0][1]) + (p[1] * R[1][1]) + (p[2] * R[2][1]);
  result[2] = (p[0] * R[0][2]) + (p[1] * R[1][2]) + (p[2] * R[2][2]);
  
  vecCopy (dest, result);
}


/*
 * vecMatch
 *
 * Input:
 *    angle
 *       Destination angle in radians.
 *    axis
 *       Destination axis.
 *    u
 *       Normalized vector.
 *    v
 *       Normalized vector.
 *
 * Description:
 *    Returns an angle R and an axis A. Rotate R radians around axis
 *    A to transform vector u to match vector v. Note that u and
 *    v must be normalized direction vectors.
 *
 */

void
vecMatch (float *angle, vector axis, vector u, vector v)
{
  vecCross (axis, u, v);
  vecNormalize (axis, axis);
  *angle = acos (vecDot (u, v));
}
